"""
Plot the convergence of eigenvalues (or corresponding frequencies) of an
eigenvalue problem to an analytical solution, when applying the uniform
mesh refinement.

Uses the PRIMME eigenvalue solver by default (``pip install primme``).

Usage Examples
--------------

- Run without the convergence analysis, use the spectral element method (SEM)
  basis of order 5::

    sfepy-run sfepy/examples/miscellaneous/refine_evp.py -d order=5,basis=sem

- Get help::

    python3 sfepy/examples/miscellaneous/refine_evp.py -h

- Plot the convergence of the smallest eigenvalue of the Laplace Dirichlet
  problem::

    python3 sfepy/examples/miscellaneous/refine_evp.py --max-order=5 --max-refine=2

- Plot the convergence of the smallest frequency of the 1D elastic bar
  vibration problem, show relative errors::

    python3 sfepy/examples/miscellaneous/refine_evp.py --max-order=5 --max-refine=2 --kind=elasticity --transform=freqs --relative

- Using the 1D elastic bar vibration problem, compare the SEM results with the
  FEM + row-sum mass matrix lumping. Plot also the sparsity patterns of the
  mass (M) and stiffness (K) matrices::

    python3 sfepy/examples/miscellaneous/refine_evp.py --max-order=5 --max-refine=2 --evps=primme --kind=elasticity-lumping --transform=freqs --relative --beta=1 --mass-lumping='row_sum' --sparsity

    python3 sfepy/examples/miscellaneous/refine_evp.py --max-order=5 --max-refine=2 --evps=primme --kind=elasticity --basis=sem --transform=freqs --relative --beta=0 --mass-lumping='none' --sparsity
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import sys
import os.path as op
from functools import partial
from itertools import product

import numpy as nm
import matplotlib.pyplot as plt

from sfepy.base.base import output
from sfepy.base.conf import dict_from_string as parse_as_dict
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.base.conf import ProblemConf
from sfepy.discrete import Problem
from sfepy.solvers import Solver
from sfepy.discrete.fem.poly_spaces import SEMTensorProductPolySpace

def define(order=1, refine=0, evps='primme', n_eigs=1, eigs_only=True,
           kind='laplace', basis='lagrange', beta=1.0, mass_lumping='none',
           transform='none', output_dir='output', **kwargs):

    options = {
        'evps' : evps,
        'n_eigs' : n_eigs,
        'eigs_only' : eigs_only,

        'active_only' : True,
        'refinement_level' : refine,

        'output_dir' : output_dir,
    }

    if kind == 'laplace':
        dims, shape, centre = [1.0, 1.0], [3, 3], [0.5, 0.5]

        regions = {
            'Omega' : 'all',
            'Gamma' : ('vertices of surface', 'facet'),
        }

        materials = {
        }

        fields = {
            'fu' : ('real', 'scalar', 'Omega', order, 'H1', basis),
        }

        variables = {
            'u' : ('unknown field', 'fu', 0),
            'v' : ('test field', 'fu', 'u'),
        }

        ebcs = {
            'zeros' : ('Gamma', {'u.0' : 0.0}),
        }

        if basis != 'sem':
            integrals = {
                'i' : 2 * order,
            }

        else:
            from sfepy.discrete.fem.geometry_element import GeometryElement
            dim = len (dims)
            gel = GeometryElement(f'{dim}_{2**dim}')
            ps = SEMTensorProductPolySpace(None, gel, order)
            integrals = {
                'i' : ('custom', ps.node_coors, ps.node_weights),
            }

        equations = {
            'lhs' : """dw_laplace.i.Omega(v, u)""",
            'rhs' : """dw_dot.i.Omega(v, u)""",
        }
        def eval_exact(n_eigs):
            num = int(nm.ceil(nm.sqrt(n_eigs)))
            rng = range(1, num+1)
            eigs = [(ir**2 + ic**2) * nm.pi**2 for ir, ic in product(rng, rng)]
            return nm.array(sorted(eigs)[:n_eigs], dtype=nm.float64)

    else:
        import sfepy.mechanics.matcoefs as mc

        dims, shape, centre = [1.0], [2], [0.5]

        regions = {
            'Omega' : 'all',
            'Gamma' : ('vertices in (x < %.16e)' % 1e-8, 'facet'),
        }
        dim = len(dims)
        young = 69e9
        density = 2700.0
        poisson = 0.0
        plane = 'strain'
        materials = {
            'm' : ({
                'rho' : density,
                'D' : mc.stiffness_from_youngpoisson(
                    dim, young=young, poisson=poisson, plane=plane
                ),
            },),
            'c' : ({
                '.beta' : beta,
                '.mass_lumping' : mass_lumping,
            },),
        }

        fields = {
            'fu' : ('real', dim, 'Omega', order, 'H1', basis),
        }

        variables = {
            'u' : ('unknown field', 'fu', 0),
            'v' : ('test field', 'fu', 'u'),
        }

        ebcs = {
            'zeros' : ('Gamma', {'u.all' : 0.0}),
        }

        if basis != 'sem':
            integrals = {
                'i' : 2 * order,
            }

        else:
            from sfepy.discrete.fem.geometry_element import GeometryElement
            dim = len(dims)
            gel = GeometryElement(f'{dim}_{2**dim}')
            ps = SEMTensorProductPolySpace(None, gel, order)

            integrals = {
                'i' : ('custom', ps.node_coors, ps.node_weights),
            }

        equations = {
            'lhs' : """dw_lin_elastic.i.Omega(m.D, v, u)""",
            'rhs' : """dw_dot.i.Omega(m.rho, v, u)""",
        }
        if kind == 'elasticity-lumping':
            equations.update({
                'rhs' : """de_mass.i.Omega(m.rho, c.mass_lumping, c.beta,
                                           v, u)""",
            })

        def eval_exact(n_eigs):
            c0 = nm.sqrt(young / density)

            ns = nm.arange(1, n_eigs+1)
            freqs = 0.5 * (ns - 0.5) * c0 / dims[0]

            if transform == 'none':
                eigs = (2 * nm.pi * freqs)**2

            else:
                eigs = freqs

            return eigs

    def mesh_hook(mesh, mode):
        """
        Generate the block mesh.
        """
        if mode == 'read':
            mesh = gen_block_mesh(dims, shape, centre,
                                  name='evp-test', verbose=False)
            return mesh

        elif mode == 'write':
            pass

    filename_mesh = UserMeshIO(mesh_hook)

    solvers = {
        'matlab': ('eig.matlab', {
            'method': 'eigs',
            'which': 'sm',
            'eps': 1e-14,
        }),
        'primme' : ('eig.primme', {
            'which' : 'SM',
             # Prevent hanging with 1x1 matrices and tol = 1e-14.
            'tol' : 1e-13 if refine == 0 else 1e-14,
        }),
        'scipy' : ('eig.scipy', {
            'method' : 'eigsh',
            'tol' : 1e-14,
            # 'maxiter' : 150,

            # Compute the eigenvalues near 0 (= the smallest eigenvalues) using
            # the shift-invert mode.
            'which' : 'LM',
            'sigma' : 0.0,
        }),
    }

    return locals()

def get_figname(name, options):
    rel = 'r' if options.relative else 'a'
    figname = (f'{name}-{options.kind}-{options.basis}-{options.evps}'
               f'-{options.transform}-{rel}{options.fig_suffix}')
    return figname

def parse_args(args=None):
    helps = {}

    opts = dict(
        max_order = (1, 'max. approximation order'),
        max_refine = (0, 'max. uniform refinement level'),
        evps = ('primme', 'solver name'),
        n_eigs = (1, 'number of eigenvalues'),
        kind = (('laplace', 'elasticity', 'elasticity-lumping'), 'problem kind'),
        basis = (('lagrange', 'lobatto', 'sem'), 'field basis'),
        beta = (1.0, 'averaged mass matrix parameter'),
        mass_lumping = ('none', 'mass matrix lumping algorithm'),
        transform = (('none', 'freqs'), 'eigenvalues transformation'),
        relative = (False, 'plot relative errors'),
        sparsity = (False, 'plot sparsity patterns of matrices'),
        fig_suffix = ('.pdf', 'file suffix for saving figures'),
        plot_rc_params = ('', 'matplotlib resources'),
        output_dir = ('output', 'output directory'),
        show = (True, 'do not show matplotlib figures'),
        shell = (False, 'run ipython shell after all computations'),
        debug = (False,
                 'automatically start debugger when an exception is raised'),
    )

    parser = ArgumentParser(description=__doc__.rstrip(),
                            formatter_class=RawDescriptionHelpFormatter)

    dhelp = ' [default: %(default)s]'
    for key, (val, msg) in opts.items():
        helps[key] = msg
        action = 'store'
        vtype = type(val)
        choices = None
        option = key
        if val is True:
            action = 'store_false'
            option = 'no_' + key

        elif val is False:
            action = 'store_true'

        elif isinstance(val, tuple):
            choices = val
            vtype = type(val[0])
            val = val[0]

        elif isinstance(val, list):
            vtype = type(val[1])
            val = val[0]

        if action == 'store':
            helps[key] += dhelp
            parser.add_argument('--' + option.replace('_', '-'),
                                type=vtype,
                                action=action, dest=key, choices=choices,
                                default=val, help=helps[key])
        else:
            parser.add_argument('--' + option.replace('_', '-'),
                                action=action, dest=key,
                                default=val, help=helps[key])
    options = parser.parse_args(args=args)

    options.plot_rc_params = parse_as_dict(options.plot_rc_params)

    return options, helps

def main():
    options, helps = parse_args()
    if options.debug:
        from sfepy.base.base import debug_on_error; debug_on_error()

    inodir = partial(op.join, options.output_dir)

    orders = range(1, options.max_order + 1)

    all_eigs = []
    for order, refine in product(
            orders,
            range(options.max_refine + 1),
    ):
        conf = ProblemConf.from_dict(define(order=order, refine=refine,
                                            **vars(options)),
                                     sys.modules[__name__])
        pb = Problem.from_conf(conf)

        pb.time_update()
        mtx_a = pb.evaluate(pb.conf.equations['lhs'], mode='weak',
                            auto_init=True, dw_mode='matrix')

        mtx_b = pb.evaluate(pb.conf.equations['rhs'], mode='weak',
                            dw_mode='matrix')
        eig = Solver.any_from_conf(pb.get_solver_conf(options.evps))

        n_eigs = min(mtx_a.shape[0], options.n_eigs)
        try:
            eigs = eig(mtx_a, mtx_b, n_eigs, eigenvectors=False)
            eigs = nm.pad(eigs, (0, options.n_eigs - n_eigs), 'constant',
                          constant_values=(nm.nan, nm.nan))

        except:
            eigs = nm.full(options.n_eigs, nm.nan)

        all_eigs.append(eigs)

    eigs_nums = nm.array(all_eigs).reshape(
        (options.max_order, -1, options.n_eigs)
    )
    eigs_ana = pb.conf.eval_exact(options.n_eigs)

    if options.transform == 'freqs':
        eigs_nums = nm.sqrt(eigs_nums) / (2 * nm.pi)

    ref = nm.arange(options.max_refine + 1)
    hs = 1.0 / 2.0**ref

    plt.rcParams['text.usetex'] = True
    plt.rcParams['font.size'] = 14
    plt.rcParams['lines.linewidth'] = 3
    plt.rcParams['legend.handlelength'] = 2.5
    plt.rcParams['legend.labelspacing'] = 0.0
    plt.rcParams.update(options.plot_rc_params)

    x = r'\lambda' if options.transform == 'none' else 'f'
    if options.relative:
        ylabel = f'$|{x}/{x}_{{ref}} - 1|$'

    else:
        ylabel = f'$|{x} - {x}_{{ref}}|$'

    for ie in range(options.n_eigs):
        fig, ax = plt.subplots()
        colors = plt.cm.tab10.colors
        for io, order in enumerate(orders):
            eigs_num = eigs_nums[io, :, ie]
            eig_ana = eigs_ana[ie]

            error = nm.abs(eigs_num - eig_ana)
            if options.relative:
                error /= eig_ana

            output('order:', order)
            output('analytical:', eig_ana)
            output('numerical :', eigs_num)

            t0 = error[0] / 2
            terror = t0 * hs**(2 * order)

            ax.semilogy(ref, error, color=colors[io], label=order)
            ax.semilogy(ref, terror, color=colors[io], ls=':')

        ax.set_xlabel('refinement level')
        ax.set_ylabel(ylabel)
        ax.legend()

        fig.tight_layout()
        fig.savefig(inodir(get_figname(f'h-refinement-{ie}', options)),
                    bbox_inches='tight')

    if options.sparsity:
        fig, ax = plt.subplots()
        ax.spy(mtx_a, marker='.', ms=4)
        fig.tight_layout()
        fig.savefig(inodir(get_figname('K-sparsity', options)),
                    bbox_inches='tight')

        fig, ax = plt.subplots()
        ax.spy(mtx_b, marker='.', ms=4)
        fig.tight_layout()
        fig.savefig(inodir(get_figname('M-sparsity', options)),
                    bbox_inches='tight')

    if options.show:
        plt.show()

    if options.shell:
        from sfepy.base.base import shell; shell()

if __name__ == '__main__':
    main()



import numpy as nm
import pytest

from sfepy.base.conf import transform_functions
import sfepy.base.testing as tst

def get_vertices(coors, domain=None):
    x, z = coors[:,0], coors[:,2]

    return nm.where((z < 0.1) & (x < 0.1))[0]

def get_cells(coors, domain=None):
    return nm.where(coors[:, 0] < 0)[0]

@pytest.fixture(scope='module')
def data():
    from sfepy import data_dir
    from sfepy.base.base import Struct
    from sfepy.discrete.fem import Mesh, FEDomain
    from sfepy.discrete import Functions

    mesh = Mesh.from_file(data_dir
                          + '/meshes/various_formats/abaqus_tet.inp')
    mesh.nodal_bcs['set0'] = [0, 7]
    domain = FEDomain('test domain', mesh)

    conf_functions = {
        'get_vertices' : (get_vertices,),
        'get_cells' : (get_cells,),
    }
    functions = Functions.from_conf(transform_functions(conf_functions))

    return Struct(domain=domain, functions=functions)

def test_selectors(data):
    """
    Test basic region selectors.
    """
    selectors = [
        ['all', 'cell'],
        ['vertices of surface', 'facet'],
        ['vertices of group %d' % data.domain.cmesh.vertex_groups[0],
         'facet'],
        ['vertices of set set0', 'vertex'],
        ['vertices in (z < 0.1) & (x < 0.1)', 'facet'],
        ['vertices by get_vertices', 'cell'],
        ['vertex 0, 1, 2', 'vertex'],
        ['vertex in r.r6', 'vertex'],
        ['cells of group %d' % data.domain.cmesh.cell_groups[0], 'cell'],
        # ['cells of set 0', 'cell'], not implemented...
        ['cells by get_cells', 'cell'],
        ['cell 1, 4, 5', 'cell'],
        ['copy r.r5', 'cell'],
        ['r.r5', 'cell'],
    ]

    vertices = [
        [0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12],
        [0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12],
        [0,  1,  3,  7],
        [0,  7],
        [1,  2,  3,  4,  5,  9, 11],
        [1,  2,  3,  4,  5,  9, 11],
        [0,  1,  2],
        [0],
        [0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10],
        [0,  1,  2,  3,  4,  5,  6,  9, 10, 11],
        [0,  1,  2,  3,  4,  5,  6,  8],
        [1,  2,  3,  4,  5,  9, 11],
        [1,  2,  3,  4,  5,  9, 11],
    ]

    data.domain.reset_regions()

    ok = True
    for ii, sel in enumerate(selectors):
        tst.report('select:', sel)
        reg = data.domain.create_region('r%d' % ii, sel[0], kind=sel[1],
                                        functions=data.functions)
        _ok = ((len(reg.vertices) == len(vertices[ii]))
               and (reg.vertices == vertices[ii]).all())
        tst.report('  vertices:', _ok)

        ok = ok and _ok

    assert ok

def test_operators(data):
    """
    Test operators in region selectors.
    """
    ok = True

    data.domain.reset_regions()

    r1 = data.domain.create_region('r1', 'all')

    sel = 'r.r1 -v vertices of group 0'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel, kind='vertex')
    av = [2,  4,  5,  6,  8,  9, 10, 11, 12]
    _ok = (reg.vertices == nm.array(av)).all()
    tst.report('  vertices:', _ok)
    ok = ok and _ok

    sel = 'vertex 0, 1, 2 +v vertices of group 0'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel, kind='vertex')
    av = [0,  1,  2,  3,  7]
    _ok = (reg.vertices == nm.array(av)).all()
    tst.report('  vertices:', _ok)
    ok = ok and _ok

    sel = 'vertex 0, 1, 2 *v vertices of group 0'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel, kind='vertex')
    av = [0,  1]
    _ok = (reg.vertices == nm.array(av)).all()
    tst.report('  vertices:', _ok)
    ok = ok and _ok

    sel = 'vertex 20 *v vertices of group 0'
    tst.report('select:', sel, ', allow_empty == False')
    _ok = False
    try:
        reg = data.domain.create_region('reg', sel, kind='vertex',
                                        allow_empty=False)
    except ValueError:
        _ok = True
    tst.report('  exception raised:', _ok)
    ok = ok and _ok

    sel = 'vertex 20 *v vertices of group 0'
    tst.report('select:', sel, ', allow_empty == True')
    reg = data.domain.create_region('reg', sel, kind='vertex',
                                    allow_empty=True)
    av = []
    _ok = (reg.vertices == nm.array(av)).all()
    tst.report('  vertices:', _ok)
    ok = ok and _ok

    sel = 'r.r1 -c cell 1, 4, 5'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel)
    _ok = (nm.setdiff1d(r1.cells[0], [1, 4, 5]) == reg.cells[0]).all()
    tst.report('  cells:', _ok)
    ok = ok and _ok

    sel = 'cell 8, 3 +c cell 1, 4, 5'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel)
    cells = [1,  3,  4,  5,  8]
    _ok = (reg.cells == nm.array(cells)).all()
    tst.report('  cells:', _ok)
    ok = ok and _ok

    sel = 'cell 8, 3, 2 *c cell 8, 4, 2, 7'
    tst.report('select:', sel)
    reg = data.domain.create_region('reg', sel)
    cells = [2,  8]
    _ok = (reg.cells == nm.array(cells)).all()
    tst.report('  cells:', _ok)
    ok = ok and _ok

    assert ok



import numpy as nm

from sfepy.base.base import get_default, Struct
from sfepy.discrete.fem.facets import build_orientation_map
from sfepy.discrete.fem.utils import prepare_remap

class FESurface(Struct):
    """Description of a surface of a finite element domain."""

    def __init__(self, name, region, efaces, volume_econn, volume_region=None):
        """nodes[leconn] == econn"""
        """nodes are sorted by node number -> same order as region.vertices"""
        self.name = get_default(name, 'surface_data_%s' % region.name)

        face_indices = region.get_facet_indices()

        faces = efaces[face_indices[:, 1]]
        if faces.size == 0 and not region.is_empty:
            raise ValueError('region with no faces! (%s)' % region.name)

        if volume_region is None:
            ii = face_indices[:, 0]
        elif hasattr(volume_region, 'get_cell_indices'):
            ii = volume_region.get_cell_indices(face_indices[:, 0])
        else:
            ii = volume_region

        try:
            ee = volume_econn[ii]

        except:
            raise ValueError('missing region face indices! (%s)'
                             % region.name)

        econn = nm.empty(faces.shape, dtype=nm.int32)
        for ir, face in enumerate(faces):
            econn[ir] = ee[ir, face]

        nodes = nm.unique(econn)
        if len(nodes):
            remap = prepare_remap(nodes, nodes.max() + 1)
            leconn = remap[econn].copy()

        else:
            leconn = econn.copy()

        n_fa, n_fp = econn.shape
        face_type = 's%d' % n_fp

        # Store bkey in SurfaceData, so that base function can be
        # queried later.
        bkey = 'b%s' % face_type[1:]

        self.econn = econn
        self.fis = nm.ascontiguousarray(face_indices.astype(nm.int32))
        self.n_fa, self.n_fp = n_fa, n_fp
        self.nodes = nodes
        self.leconn = leconn
        self.face_type = face_type
        self.bkey = bkey
        self.meconn = {}
        self.mleconn = {}
        self.set_orientation_map()

    @staticmethod
    def from_region(name, region, ret_gel=False):
        face_indices = region.get_facet_indices()
        cells = face_indices[:, 0]

        econn, gel = region.domain.get_conn(ret_gel=True,
                                            tdim=region.tdim, cells=cells)

        surface = FESurface(name, region, gel.get_surface_entities(), econn,
                            slice(0, econn.shape[0]))

        if ret_gel:
            return surface, gel.surface_facet
        else:
            return surface

    def set_orientation_map(self):
        n_fp = self.n_fp
        if n_fp <= 4:
            oo, _, _ = build_orientation_map(n_fp)
            ori_map = nm.zeros((nm.max(list(oo.keys())) + 1, n_fp),
                               dtype=nm.int32)
            ori_map[list(oo.keys())] = nm.array([ii[1] for ii in oo.values()])
            self.ori_map = ori_map
        else:
            self.ori_map = None

    def setup_mirror_connectivity(self, region, mirror_name):
        """
        Setup mirror surface connectivity required to integrate over a
        mirror region.

        1. Get orientation of the faces:
           a) for own elements -> ooris
           b) for mirror elements -> moris

        2. orientation -> permutation.
        """
        def get_omap(reg, ori_map):
            if reg.tdim == (reg.dim - 1):
                cells = reg.get_cells()
                conn = reg.domain.get_conn(tdim=reg.tdim)[cells, :]
                return nm.argsort(conn)
            else:
                conn = reg.cmesh.get_conn_as_graph(reg.dim, reg.dim - 1)
                oris = reg.cmesh.facet_oris
                fis = reg.get_facet_indices()
                return ori_map[oris[conn.indptr[fis[:, 0]] + fis[:, 1]]]

        if mirror_name in self.meconn:
            return

        mregion = region.get_mirror_region(mirror_name)
        mmap = get_omap(mregion, self.ori_map)
        omap = get_omap(region, self.ori_map)

        econn = self.econn
        if mirror_name in region.mirror_maps\
            and region.mirror_maps[mirror_name] is not None:
            mirror_map = region.mirror_maps[mirror_name]
            omap = omap[mirror_map]
            econn = econn[mirror_map]

        n_el, n_ep = econn.shape
        ii = nm.repeat(nm.arange(n_el)[:, None], n_ep, 1)
        meconn = nm.empty_like(econn)
        meconn[ii, mmap] = econn[ii, omap]

        nodes = nm.unique(meconn)
        remap = prepare_remap(nodes, nodes.max() + 1)
        self.meconn[mirror_name] = meconn
        self.mleconn[mirror_name] = remap[meconn].copy()

    def get_connectivity(self, local=False, trace_region=None):
        """
        Return the surface element connectivity.

        Parameters
        ----------
        local : bool
            If True, return local connectivity w.r.t. surface nodes,
            otherwise return global connectivity w.r.t. all mesh nodes.
        trace_trace : None or str
            If not None, return mirror connectivity according to `local`.
        """
        if trace_region is None:
            if local:
                return self.leconn

            else:
                return self.econn

        else:
            if local:
                return self.mleconn[trace_region]

            else:
                return self.meconn[trace_region]


class FEPhantomSurface(FESurface):
    """A phantom surface of the region with tdim=2."""

    def __init__(self, name, region, volume_econn):
        self.name = get_default(name, 'surface_data_%s' % region.name)

        ii = region.get_cells()
        econn = volume_econn[ii]
        nodes = nm.unique(econn)

        if len(nodes):
            remap = prepare_remap(nodes, nodes.max() + 1)
            leconn = remap[econn].copy()

        else:
            leconn = econn.copy()

        n_fa, n_fp = econn.shape
        face_type = 's%d' % n_fp

        # Store bkey in SurfaceData, so that base function can be
        # queried later.
        bkey = 'b%s' % face_type[1:]

        self.econn = econn
        self.fis = -nm.ones((n_fa, 2), dtype=nm.int32)
        self.fis[:, 0] = ii
        self.n_fa, self.n_fp = n_fa, n_fp
        self.nodes = nodes
        self.leconn = leconn
        self.face_type = face_type
        self.bkey = bkey
        self.meconn = {}
        self.mleconn = {}
        self.set_orientation_map()



r"""
A linear elastic beam loaded with a continuous force. The FE mesh consists
of hexehedral and tetrahedral elements.

The displacement at the beam end is compared to the reference
solution calculated on the homogeneous hexahedral mesh.

Running the simulation::

    sfepy-run sfepy/examples/linear_elasticity/mixed_mesh.py

Viewing the results::

    sfepy-view beam_h* -f u:s0:wu:e:p0 u:s1:wu:e:p0 --camera-position="1.2,-0.6,0.1,0.4,0.1,-0.1,-0.2,0.1,1"
"""
from __future__ import absolute_import
from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy.base.base import Struct
import numpy as nm


def reference_solution(pb):
    from sfepy.discrete import Problem

    conf = pb.conf.copy()
    conf.filename_mesh = data_dir + '/meshes/3d/beam_h7.mesh'
    for k in list(conf.regions.keys()):
        if '_t' in k:
            del conf.regions[k]

    for k in list(conf.fields.keys()):
        if '_t' in k:
            del conf.fields[k]

    for k in list(conf.variables.keys()):
        if '_t' in k:
            del conf.variables[k]

    conf.equations = {
        'balance_of_forces':
            """dw_lin_elastic.i.Omega(solid.D, v, u)
             = dw_surface_ltr.i.Top(force.val, v)"""
    }

    rpb = Problem.from_conf(conf)

    return rpb.solve().get_state_parts()['u']


def get_force(ts, coors, mode=None, **kwargs):
    if mode == 'qp':
        F = 1e3

        val = nm.zeros_like(coors)[..., None]
        val[:, 2, 0] = -coors[:, 0] / 0.7 * F

        return {'val': val}


def post_proces(out, pb, state, extend=False):
    c1 = pb.domain.regions['Omega'].cells
    c2 = pb.domain.regions['Omega_t'].cells
    ed = pb.domain.regions['Edge'].vertices

    S = nm.zeros((pb.domain.cmesh.n_el, 1, 6, 1), dtype=nm.float64)

    S[c1] = pb.evaluate('ev_cauchy_stress.i.Omega(solid.D, u)',
                        mode='el_avg')
    S[c2] = pb.evaluate('ev_cauchy_stress.i.Omega_t(solid.D, u_t)',
                        mode='el_avg')

    out['stress'] = Struct(name='out', data=S, mode='cell')

    u_nd = out['u'].data[ed[0], :]
    ref_u_nd = reference_solution(pb).reshape(-1, 3)[ed[0], :]

    du = nm.linalg.norm((ref_u_nd - u_nd) / nm.linalg.norm(ref_u_nd))
    print(f'Relative difference with respect to the reference solution: {du}')

    return out


filename_mesh = data_dir + '/meshes/3d/beam_h5t12.mesh'

options = {
    'post_process_hook': 'post_proces',
}

regions = {
    'Omega_': 'all',
    'Omega': ('cells of group 1', 'cell', None, {'vertices_from': 'Omega_'}),
    'Omega_t': ('cells of group 2', 'cell', None, {'vertices_from': 'Omega_'}),
    'Left': ('vertices in (x < 0.01)', 'facet'),
    'Right': ('vertices in (x > 0.69)', 'facet'),
    'Top_': ('vertices in (z > 0.09)', 'facet'),
    'Bottom': ('vertices in (z < 0.01)', 'facet'),
    'Edge': ('r.Bottom *v r.Right', 'vertex'),
    'Top': ('r.Top_ *f r.Omega', 'facet'),
    'Top_t': ('r.Top_ *f r.Omega_t', 'facet'),
}

functions = {
    'get_force' : (get_force,),
}

materials = {
    'solid': ({'D': stiffness_from_youngpoisson(dim=3, young=1e6, poisson=0.3)},),
    'force': 'get_force',
    'vforce': ({'.val' : [0.0, 0.0, -7]},),
}

fields = {
    'displacement': ('real', 'vector', 'Omega', 1),
    'displacement_t': ('real', 'vector', 'Omega_t', 1),
}

integrals = {
    'i': 2,
}

variables = {
    'u': ('unknown field', 'displacement', 0),
    'v': ('test field', 'displacement', 'u'),
    'u_t': ('unknown field', 'displacement_t', 0),
    'v_t': ('test field', 'displacement_t', 'u_t'),
}

ebcs = {
    'Fixed': ('Left', {'u.all' : 0.0}),
}

equations = {
    'balance_of_forces':
    """dw_lin_elastic.i.Omega(solid.D, v, u)
     + dw_lin_elastic.i.Omega_t(solid.D, v_t, u_t)
     = dw_surface_ltr.i.Top(force.val, v)
     + dw_surface_ltr.i.Top_t(force.val, v_t)"""
}

solvers = {
    'ls': ('ls.auto_direct', {}),
    'newton': ('nls.newton', {
        'i_max': 1,
        'eps_a': 1e-6,
    }),
}



# -*- Mode: Python -*-
"""
Low level reference mapping functionality.
"""
cimport cython

from _fmfield cimport (FMField, array2fmfield4)

from types cimport int32, float64

cdef extern from 'refmaps.h':

    ctypedef struct Mapping:
        int32 nEl
        int32 nQP
        int32 dim
        int32 nEP
        FMField *bf
        FMField *bfGM # Volume or SurfaceExtra only.
        FMField *det # detJMR or detJSR.
        FMField *normal # Surface only.
        FMField *volume
        float64 totalVolume

cdef class CMapping:
    cdef Mapping geo[1]
    cdef FMField[1] _bf, _bfg, _det, _normal, _volume



r"""
An example demonstrating the usage of the truss structural elements in 3D.
The mixed (solid and structural elements) bridge structure is fixed
on the left and supported on the right.

Running the simulation::

  sfepy-run sfepy/examples/linear_elasticity/truss_bridge3d.py

Viewing the results::

  sfepy-view bridge3d_S*.vtk -f u_solid:s0:wu_solid:f1e3:p0 u_struct:s1:wu_struct:f1e3:p0
"""
import numpy as nm
from sfepy.base.base import Struct
from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy import data_dir


def post_process(out, pb, state, extend=False):
    valP = pb.evaluate('ev_lin_truss_force.0.Struct(truss.EA, u)',
                       mode='el_avg')
    out['P'] = Struct(name='output_data', mode='cell',
                      region_name='Struct', data=valP)

    valS = pb.evaluate('ev_cauchy_stress.i.Solid(solid.D, u)', mode='el_avg')
    out['S'] = Struct(name='output_data', mode='cell',
                      region_name='Solid', data=valS)

    u = out.pop('u').data
    u_solid = u[pb.domain.regions['Solid'].vertices, :]
    out['u_solid'] = Struct(name='output_data', mode='vertex',
                            region_name='Solid', data=u_solid)
    u_truss = u[pb.domain.regions['Struct'].vertices, :]
    out['u_struct'] = Struct(name='output_data', mode='vertex',
                             region_name='Struct', data=u_truss)

    return out


filename_mesh = data_dir + "/meshes/3d/bridge3d.vtk"

options = {
    'post_process_hook': 'post_process',
    'split_results_by': 'region',
}

regions = {
    'Solid': 'cells of group 1',
    'Struct': 'cells of group 2',
    'Omega': ('r.Solid +v r.Struct', 'cell', None, {'finalize': False}),
    'Left': ('vertices in (x < 0.01)', 'vertex'),
    'Right': ('vertices in (x > 11.99)', 'vertex'),
    'Bottom': ('vertices in (z < -0.199)', 'vertex'),
    'Top': ('vertices in (z > -0.001)', 'facet'),
    'LeftBottom': ('r.Left *v r.Bottom', 'vertex'),
    'RightBottom': ('r.Right *v r.Bottom', 'vertex'),
}

materials = {
    'force': ({'val': nm.array([[0, 0, -1000.]]).T},),
    'truss': ({'EA': 210e9 * (0.02**2 * nm.pi)},),
    'solid': ({'D': stiffness_from_youngpoisson(3, 10e9, 0.3)},),
}

fields = {
    'displacement': ('real', 'vector', 'Omega', 1),
}

variables = {
    'u': ('unknown field', 'displacement'),
    'v': ('test field', 'displacement', 'u'),
}

integrals = {
    'i': 2,
}

ebcs = {
    'Fixed_Left': ('LeftBottom', {'u.all': 0.0}),
    'Support_Right': ('RightBottom', {'u.2': 0.0}),
}

equations = {
    'balance_of_forces':
    """dw_lin_elastic.i.Solid(solid.D, v, u)
     + dw_lin_truss.0.Struct(truss.EA, v, u)
     = dw_surface_ltr.i.Top(force.val, v)""",
}

solvers = {
    'ls': ('ls.auto_direct', {}),
    'newton': ('nls.newton', {'eps_a': 1e-6, 'eps_r': 1e-9}),
}



#!/usr/bin/env python
"""
Run SfePy tests. All arguments are passed to pytest.
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter

def main():
    parser = ArgumentParser(description=__doc__,
                            usage='sfepy-test [-h] [pytest options]',
                            formatter_class=RawDescriptionHelpFormatter)
    _, pytest_options = parser.parse_known_args()

    import sfepy
    sfepy.test(*pytest_options)

if __name__ == '__main__':
    main()



import numpy as nm
import pytest

from sfepy import data_dir
from sfepy.base.base import assert_
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/2d/square_unit_tri.mesh'

def get_pars(ts, coors, mode=None, extra_arg=None,
             equations=None, term=None, problem=None, **kwargs):
    if mode == 'special':
        if extra_arg == 'hello!':
            ic = 0
        else:
            ic = 1
        coors = problem.get_mesh_coors()
        return {('x_%s' % ic) : coors[:,ic]}

    elif mode == 'qp':
        return {'a' : nm.tile(-2 + 1j, (coors.shape[0], 1, 1))}

def get_p_edge(ts, coors, bc=None, **kwargs):
    if bc.name == 'p_left':
        return nm.sin(nm.pi * coors[:,1])
    else:
        return nm.cos(nm.pi * coors[:,1])

def get_u_edge(ts, coors, bc=None, **kwargs):
    out = nm.zeros_like(coors)
    out[:, 1] = nm.arange(out.shape[0]) + 1.0
    return out

def get_circle(coors, domain=None):
    r = nm.sqrt(coors[:,0]**2.0 + coors[:,1]**2.0)
    return nm.where(r < 0.2)[0]

functions = {
    'get_pars' : (get_pars,),
    'get_pars1' : (lambda ts, coors, mode=None, **kwargs:
                   get_pars(ts, coors, mode, extra_arg='hello!', **kwargs),),
    'get_p_edge' : (get_p_edge,),
    'get_u_edge' : (get_u_edge,),
    'get_circle' : (get_circle,),
}

# Just another way of adding a function, besides 'functions' keyword.
function_1 = {
    'name' : 'get_pars2',
    'function' : lambda ts, coors, mode=None, **kwargs:
        get_pars(ts, coors, mode, extra_arg='hi!', **kwargs),
}

materials = {
    'mf1' : (None, 'get_pars1'),
    'mf2' : 'get_pars2',
    # Dot denotes a special value, that is not propagated to all QP.
    'mf3' : ({'a' : 10.0, 'b' : 2.0, '.c' : 'ahoj'},),
    # Complex values.
    'mf4' : 'get_pars',
    'mf5' : ({'a' : -2 - 1j},),
    'mf6' : ({'a' : {'Circle' : 1 + 1j, 'Rest' : 3j}},),
}

fields = {
    'pressure' : (nm.float64, 1, 'Omega', 2),
    'displacement' : (nm.float64, 2, 'Omega', 2),
    'complex' : (nm.complex128, 1, 'Omega', 2),
}

variables = {
    'p'   : ('unknown field', 'pressure', 0),
    'q'   : ('test field',    'pressure', 'p'),
    'u'   : ('unknown field', 'displacement', 1),
    'v'   : ('test field',    'displacement', 'u'),
}

wx = 0.499
regions = {
    'Omega' : 'all',
    'Left' : ('vertices in (x < -%.3f)' % wx, 'facet'),
    'Right' : ('vertices in (x > %.3f)' % wx, 'facet'),
    'Circle' : 'vertices by get_circle',
    'Rest' : 'r.Omega -c r.Circle',
}

ebcs = {
    'p_left' : ('Left', {'p.all' : 'get_p_edge'}),
    'p_right' : ('Right', {'p.all' : 'get_p_edge'}),
    'u_right' : ('Right', {'u.all' : 'get_u_edge'}),
}

equations = {
    'e1' : """dw_laplace.2.Omega( mf3.a, q, p ) = 0""",
    'e2' : """dw_div_grad.2.Omega( mf3.b, v, u ) = 0""",
}

variables2 = {
    'r'   : ('unknown field', 'complex', 2),
    's'   : ('test field',    'complex', 'r'),
}

equations2 = {
    'e3' : """dw_laplace.2.Omega(mf4.a, s, r) = 0""",
    'e4' : """dw_laplace.2.Omega(mf5.a, s, r) = 0""",
    'e5' : """dw_laplace.2.Circle(mf6.a, s, r)
            + dw_laplace.2.Rest(mf6.a, s, r) = 0""",
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',
}

@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    problem = Problem.from_conf(conf)
    return problem

def test_material_functions(problem):
    from sfepy.discrete import Material
    from sfepy.base.conf import transform_variables

    conf = problem.conf

    ts = problem.get_default_ts(step=0)

    conf_mat1 = conf.get_item_by_name('materials', 'mf1')
    mat1 = Material.from_conf(conf_mat1, problem.functions)
    mat1.time_update(ts, None, mode='normal', problem=problem)

    coors = problem.domain.get_mesh_coors()
    assert_(nm.all(coors[:,0] == mat1.get_data(None, 'x_0')))

    conf_mat2 = conf.get_item_by_name('materials', 'mf2')
    mat2 = Material.from_conf(conf_mat2, problem.functions)
    mat2.time_update(ts, None, mode='normal', problem=problem)

    assert_(nm.all(coors[:,1] == mat2.get_data(None, 'x_1')))

    materials = problem.get_materials()
    materials.time_update(ts, problem.equations, mode='normal',
                          problem=problem)
    mat3 = materials['mf3']
    key = mat3.get_keys(region_name='Omega')[0]

    assert_(nm.all(mat3.get_data(key, 'a') == 10.0))
    assert_(nm.all(mat3.get_data(key, 'b') == 2.0))
    assert_(mat3.get_data(None, 'c') == 'ahoj')

    pb = problem.copy()
    pb.set_variables(transform_variables(conf.variables2))
    pb.set_equations(conf.equations2)
    materials = pb.get_materials()
    materials.time_update(ts, pb.equations, mode='normal', problem=pb)

    mat4 = materials['mf4']
    key = mat4.get_keys(region_name='Omega')[0]
    assert_(nm.all(mat4.get_data(key, 'a') == -2 + 1j))

    mat5 = materials['mf5']
    key = mat5.get_keys(region_name='Omega')[0]
    assert_(nm.all(mat5.get_data(key, 'a') == -2 - 1j))

    mat6 = materials['mf6']
    key = mat6.get_keys(region_name='Circle')[0]
    assert_(nm.all(mat6.get_data(key, 'a') == 1 + 1j))
    key = mat6.get_keys(region_name='Rest')[0]
    assert_(nm.all(mat6.get_data(key, 'a') == 3j))

def test_ebc_functions(problem, output_dir):
    import os.path as op

    problem.set_equations(problem.conf.equations)

    problem.time_update()
    state = problem.solve(save_results=False)
    name = op.join(output_dir, 'test_ebc_functions.vtk')
    problem.save_state(name, state)

    ok = True
    domain = problem.domain

    vecs = state.get_state_parts()
    vec = vecs['p']

    iv = domain.regions['Left'].vertices
    coors = domain.get_mesh_coors()[iv]
    _ok = tst.compare_vectors(vec[iv], nm.sin(nm.pi * coors[:,1]),
                              label1='p_state_left',
                              label2='p_bc_left')
    ok = _ok and ok

    iv = domain.regions['Right'].vertices
    coors = domain.get_mesh_coors()[iv]
    _ok = tst.compare_vectors(vec[iv], nm.cos(nm.pi * coors[:,1]),
                              label1='p_state_right',
                              label2='p_bc_right')
    ok = _ok and ok

    vec = vecs['u']
    vec.shape = (-1, 2)
    ok = tst.compare_vectors(vec[iv, 0], nm.zeros(len(iv)),
                             label1='u_0_state_right',
                             label2='u_0_bc_right')
    ok = _ok and ok

    ok = tst.compare_vectors(vec[iv, 1], nm.arange(len(iv)) + 1.0,
                             label1='u_1_state_right',
                             label2='u_1_bc_right')
    ok = _ok and ok

    assert ok

def test_region_functions(problem, output_dir):
    import os.path as op

    name = op.join(output_dir, 'test_region_functions.vtk')
    problem.save_regions_as_groups(name, ['Circle'])



# -*- Mode: Python -*-
"""
Low level reference mapping functionality.
"""

cdef class CMapping:
    def __cinit__(self, bf, det, volume, bfg, normal, dim):
        array2fmfield4(self._bf, bf)
        self.geo.bf = self._bf

        array2fmfield4(self._det, det)
        self.geo.det = self._det

        array2fmfield4(self._volume, volume)
        self.geo.volume = self._volume

        if bfg is not None:
            array2fmfield4(self._bfg, bfg)
            self.geo.bfGM = self._bfg
        else:
            self.geo.bfGM = NULL

        if normal is not None:
            array2fmfield4(self._normal, normal)
            self.geo.normal = self._normal
        else:
            self.geo.normal = NULL

        n_el, n_qp = det.shape[:2]
        n_ep = bf.shape[3]
        self.geo.nEl = n_el
        self.geo.nQP = n_qp
        self.geo.dim = dim
        self.geo.nEP = n_ep

    def __str__(self):
        return 'CMapping: n_el %d, n_qp %d, dim: %d, n_ep: %d' \
               % (self.geo.nEl, self.geo.nQP, self.geo.dim, self.geo.nEP)



import os.path as op
import numpy as nm
import pytest

import sfepy.base.testing as tst

def test_stokes_slip_bc(output_dir):
    import scipy.sparse as sp

    from sfepy.base.conf import ProblemConf
    from sfepy.discrete import Problem
    import sfepy.examples.navier_stokes.stokes_slip_bc as ssb

    conf = ProblemConf.from_dict(
        ssb.define(output_dir=output_dir, save_lcbc_vecs=True),
        ssb,
    )
    pb = Problem.from_conf(conf, init_solvers=False)

    pb.time_update()
    variables = pb.get_variables()

    adi = variables.adi
    lcdi = variables.lcdi
    mtx = variables.mtx_lcbc

    ok = adi.var_names == lcdi.var_names
    tst.report('same adi-lcdi ordering:', ok)

    ublock = mtx[adi.indx['u']]
    ir, ic = ublock.nonzero()
    ir += adi.indx['u'].start

    i0, i1 = adi.indx['u'].start, adi.indx['u'].stop
    _ok0 = (i0 <= ir).all() and  (ir < i1).all()
    tst.report('u block rows in [%d %d[: %s' % (i0, i1, _ok0))

    i0, i1 = lcdi.indx['u'].start, lcdi.indx['u'].stop
    _ok1 = (i0 <= ic).all() and  (ic < i1).all()
    tst.report('u block cols in [%d %d[: %s' % (i0, i1, _ok1))

    ok = ok and _ok0 and _ok1

    pblock = mtx[adi.indx['p']]
    ir, ic, iv = sp.find(pblock)
    ir += adi.indx['p'].start

    i0, i1 = adi.indx['p'].start, adi.indx['p'].stop
    _ok0 = (i0 <= ir).all() and  (ir < i1).all()
    tst.report('p block rows in [%d %d[: %s' % (i0, i1, _ok0))

    i0, i1 = lcdi.indx['p'].start, lcdi.indx['p'].stop
    _ok1 = (i0 <= ic).all() and  (ic < i1).all()
    tst.report('p block cols in [%d %d[: %s' % (i0, i1, _ok1))

    ok = ok and _ok0 and _ok1

    _ok0 = (len(ir) == adi.n_dof['p'])
    tst.report('p block size correct:', _ok0)
    _ok1 = ((ir - adi.indx['p'].start) == (ic - lcdi.indx['p'].start)).all()
    tst.report('p block diagonal:', _ok1)
    _ok2 = (iv == 1.0).all()
    tst.report('p block identity:', _ok2)

    ok = ok and _ok0 and _ok1 and _ok2

    assert ok

def test_laplace_shifted_periodic(output_dir):
    from sfepy.mesh.mesh_generators import gen_block_mesh
    from sfepy.discrete.fem import FEDomain
    from sfepy.examples.diffusion.laplace_shifted_periodic import run

    dims = [2.0, 1.0]
    shape = [21, 11]
    centre = [0.0, 0.0]
    mesh = gen_block_mesh(dims, shape, centre, name='block-fem')
    domain = FEDomain('test_laplace_shifted_periodic', mesh)

    pb, state = run(domain, 3, output_dir=output_dir)

    gamma3 = pb.domain.regions['Gamma3']
    gamma4 = pb.domain.regions['Gamma4']

    field = pb.fields['fu']

    # Check that the shift equals to one.
    i3 = field.get_dofs_in_region(gamma3, merge=True)
    i4 = field.get_dofs_in_region(gamma4, merge=True)

    i_corners = nm.array([0, shape[0] - 1])
    ii = nm.setdiff1d(nm.arange(len(i3)), i_corners)

    vals = state()

    shift = vals[i3] - vals[i4]

    ok = (shift[i_corners] == 0.0).all()

    ok = ok and nm.allclose(shift[ii], 1.0, rtol=0.0, atol=1e-14)

    if not ok:
        tst.report('wrong shift:', shift)

    assert ok

@pytest.mark.parametrize('mesh_filename', [
    'meshes/2d/special/circle_in_square.mesh',
    'meshes/3d/special/cube_sphere.mesh',
])
def test_elasticity_rigid(mesh_filename, output_dir):
    from sfepy import data_dir
    from sfepy.base.base import IndexedStruct
    from sfepy.discrete import (FieldVariable, Material, Integral,
                                Equation, Equations, Problem)
    from sfepy.discrete.fem import Mesh, FEDomain, Field
    from sfepy.mechanics.matcoefs import stiffness_from_lame
    from sfepy.terms import Term
    from sfepy.discrete.conditions import (Conditions, EssentialBC,
                                           LinearCombinationBC)
    from sfepy.solvers.ls import ScipyDirect
    from sfepy.solvers.nls import Newton

    filename = op.join(data_dir, mesh_filename)
    mesh = Mesh.from_file(filename)
    domain = FEDomain('domain', mesh)

    min_x, max_x = domain.get_mesh_bounding_box()[:,0]
    eps = 1e-8 * (max_x - min_x)

    axis = {2 : 'y', 3 : 'z'}[mesh.dim]
    order = {2 : 2, 3 : 1}[mesh.dim]

    omega = domain.create_region('Omega', 'all')
    yrig = domain.create_region('Yrig', 'cells of group 2')
    bottom = domain.create_region('Bottom',
                                  f'vertices in {axis} < {min_x + eps}',
                                  'facet')
    top = domain.create_region('Top',
                               f'vertices in {axis} > {max_x - eps}',
                               'facet')
    field = Field.from_args('fu', nm.float64, 'vector', omega,
                            approx_order=order)
    u = FieldVariable('u', 'unknown', field)
    v = FieldVariable('v', 'test', field, primary_var_name='u')

    m = Material('m', D=stiffness_from_lame(dim=mesh.dim, lam=10.0, mu=1.0))

    integral = Integral('i', order=2 * order)
    t1 = Term.new('dw_lin_elastic(m.D, v, u)',
                  integral, omega, m=m, v=v, u=u)
    eq = Equation('balance', t1)
    eqs = Equations([eq])

    fix = EssentialBC('fix', bottom, {'u.all' : 0.0})
    load = EssentialBC('load', top, {'u.all' : 0.3})
    rig = LinearCombinationBC('rig', [yrig, None], {'u.all' : None},
                              None, 'rigid')

    ls = ScipyDirect({})

    nls_status = IndexedStruct()
    nls = Newton({}, lin_solver=ls, status=nls_status)

    pb = Problem('elasticity', equations=eqs)

    trunk = f'test_elasticity_rigid_{mesh.dim}d'
    pb.setup_output(output_filename_trunk=trunk, output_dir=output_dir)
    pb.save_regions_as_groups(op.join(output_dir, trunk + '_regions'))

    pb.set_bcs(ebcs=Conditions([fix, load]), lcbcs=Conditions([rig]))
    pb.set_solver(nls)

    status = IndexedStruct()
    pb.solve(status=status)

    assert nls_status.condition == 0

    strain = u.evaluate('cauchy_strain', region=yrig, integral=integral)

    mstrain = nm.abs(strain).max()
    tst.report('max |strain| in rigid region:', mstrain)

    assert mstrain < 1e-13



import pytest

from sfepy import data_dir
from sfepy.mechanics.matcoefs import stiffness_from_lame
import sfepy.base.testing as tst

filename_meshes = ['/meshes/3d/cube_medium_tetra.mesh',
                   '/meshes/3d/cube_medium_tetra.mesh',
                   '/meshes/3d/cube_medium_hexa.mesh']
filename_meshes = [data_dir + name for name in filename_meshes]

all_your_bases = [1, 2, 1]

filename_mesh = None

field_1 = {
    'name' : '3_displacement',
    'dtype' : 'real',
    'shape' : (3,),
    'region' : 'Omega',
    'approx_order' : None,
}

def get_pars( dim, full = False ):
    import numpy as nm
    sym = (dim + 1) * dim // 2
    lam = 1e1
    mu = 1e0
    o = nm.array( [1.] * dim + [0.] * (sym - dim), dtype = nm.float64 )
    oot = nm.outer( o, o )
    if full:
        return lam * oot + mu * nm.diag( o + 1.0 )
    else:
        return lam, mu

material_1 = {
    'name' : 'solid',
    'values' : {
        'Dijkl' : get_pars( 3, True ),
        'D' : stiffness_from_lame(3, get_pars(3)[0], get_pars(3)[1]),
        'lam' : get_pars(3)[0],
        'mu' : get_pars(3)[1],
    }
}

material_2 = {
    'name' : 'spring',
    'values' : {
        '.stiffness' : 1e0,
    }
}

variable_1 = {
    'name' : 'u',
    'kind' : 'unknown field',
    'field' : '3_displacement',
    'order' : 0,
}
variable_2 = {
    'name' : 'v',
    'kind' : 'test field',
    'field' : '3_displacement',
    'dual' : 'u',
}

region_1000 = {
    'name' : 'Omega',
    'select' : 'all',
}

region_1 = {
    'name' : 'Bottom',
    'select' : 'vertices in (z < -0.499)',
    'kind' : 'facet',
}
region_2 = {
    'name' : 'Top',
    'select' : 'vertices in (z > 0.499)',
    'kind' : 'facet',
}

ebc_1 = {
    'name' : 'Load',
    'region' : 'Top',
    'dofs' : {'u.2' : 0.1},
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

equations_getpars = {
    'balance_of_forces' :
    """dw_lin_elastic.i.Omega(solid.Dijkl, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

equations_matcoefs = {
    'balance_of_forces' :
    """dw_lin_elastic.i.Omega(solid.D, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

equations_iso = {
    'balance_of_forces' :
    """dw_lin_elastic_iso.i.Omega(solid.lam, solid.mu, v, u)
     = dw_point_lspring.i.Bottom(spring.stiffness, v, u)""",
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',
    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

@pytest.fixture(scope='module')
def solutions(output_dir):
    import sys
    from sfepy.applications import solve_pde
    from sfepy.base.base import IndexedStruct, Struct
    from sfepy.base.conf import ProblemConf
    import os.path as op

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    solutions = []
    for ii, approx_order in enumerate(all_your_bases):
        sols = Struct()

        fname = filename_meshes[ii]

        conf.filename_mesh = fname
        fields = {'field_1' : {
                      'name' : '3_displacement',
                      'dtype' : 'real',
                      'shape' : (3,),
                      'region' : 'Omega',
                      'approx_order' : approx_order,
                }
        }
        conf.edit('fields', fields)
        tst.report('mesh: %s, base: %s' % (fname, approx_order))
        status = IndexedStruct()

        tst.report('getpars')
        conf.equations = conf.equations_getpars
        problem, state1 = solve_pde(conf, status=status,
                                    save_results=False)
        converged = status.nls_status.condition == 0
        sols.getpars = (converged, state1().copy())
        tst.report('converged: %s' % converged)

        tst.report('matcoefs')
        conf.equations = conf.equations_matcoefs
        problem, state2 = solve_pde(conf, status=status,
                                    save_results=False)
        converged = status.nls_status.condition == 0
        sols.matcoefs = (converged, state2().copy())
        tst.report('converged: %s' % converged)

        tst.report('iso')
        conf.equations = conf.equations_iso
        problem, state3 = solve_pde(conf, status=status,
                                    save_results=False)
        converged = status.nls_status.condition == 0
        sols.iso = (converged, state3().copy())
        tst.report('converged: %s' % converged)

        solutions.append(sols)

        name = op.join(output_dir,
                       '_'.join(('test_elasticity_small_strain',
                                 op.splitext(op.basename(fname))[0],
                                 '%d' % approx_order))
                       + '.vtk')
        problem.save_state(name, state1)

    return solutions

def test_converged(solutions):
    for sols in solutions:
        assert sols.getpars[0]
        assert sols.matcoefs[0]
        assert sols.iso[0]

def test_linear_terms(solutions):
    for sols in solutions:
        assert tst.compare_vectors(sols.getpars[1], sols.matcoefs[1],
                                   label1='getpars',
                                   label2='matcoefs')
        assert tst.compare_vectors(sols.getpars[1], sols.iso[1],
                                   label1='getpars',
                                   label2='iso')



#!/usr/bin/env python
# 12.01.2007, c
"""
Solve partial differential equations given in a SfePy problem definition file.

Example problem definition files can be found in ``sfepy/examples/`` directory
of the SfePy top-level directory.

In the examples below it is supposed that sfepy is installed. When using the
in-place build, replace ``sfepy-run`` by ``python3 sfepy/scripts/simple.py``.

The supported application kinds (--app option) are:

- bvp - boundary value problem. Example::

    sfepy-run sfepy/examples/diffusion/poisson.py

- homogen - calculation of local microscopic problems (correctors) and
  homogenized coefficients. Example::

    sfepy-run sfepy/examples/homogenization/perfusion_micro.py

- bvp-mM - micro-macro boundary value problem. Solve a coupled two-scale
  problem in parallel using MPI. One computational node is solving a
  macroscopic equation while the others are solving local microscopic problems
  and homogenized coefficients. The --app option is required in this case.
  Example::

    mpiexec -n 4 sfepy-run --app=bvp-mM --debug-mpi sfepy/examples/homogenization/nonlinear_hyperelastic_mM.py

- evp - eigenvalue problem. Example::

    sfepy-run sfepy/examples/quantum/well.py

- phonon - phononic band gaps. Example::

    sfepy-run sfepy/examples/phononic/band_gaps.py --phonon-plot

Both normal and parametric study runs are supported. A parametric study allows
repeated runs for varying some of the simulation parameters - see
``sfepy/examples/diffusion/poisson_parametric_study.py`` file.
"""
from __future__ import print_function
from __future__ import absolute_import
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import sfepy
from sfepy.base.base import output, Struct
from sfepy.base.conf import ProblemConf, get_standard_keywords
from sfepy.applications import PDESolverApp, EVPSolverApp

def print_terms():
    import sfepy.terms as t
    tt = t.term_table
    print('Terms: %d available:' % len(tt))
    print(sorted(tt.keys()))

def print_solvers():
    from sfepy.solvers import solver_table
    print('Solvers: %d available:' % len(solver_table))
    print(sorted(solver_table.keys()))

helps = {
    'app' :
    'override application kind, normally determined automatically.'
    ' The supported kinds are:'
    ' bvp (boundary value problem),'
    ' homogen (correctors, homogenized coefficients),'
    ' bvp-mM (micro-macro boundary value problem,'
    '         homogenized coefficients computed in parallel using MPI),'
    ' evp (eigenvalue problem),'
    ' phonon (phononic band gaps)',
    'debug':
    'automatically start debugger when an exception is raised',
    'debug_mpi': 'log MPI communication (mM mode only)',
    'conf' :
    'override problem description file items, written as python'
    ' dictionary without surrounding braces',
    'options' : 'override options item of problem description,'
    ' written as python dictionary without surrounding braces',
    'define' : 'pass given arguments written as python dictionary'
    ' without surrounding braces to define() function of problem description'
    ' file',
    'filename' :
    'basename of output file(s) [default: <basename of input file>]',
    'output_format' :
    'output file format, one of: {vtk, h5} [default: vtk]',
    'save_restart' :
    'if given, save restart files according to the given mode.',
    'load_restart' :
    'if given, load the given restart file',
    'log' :
    'log all messages to specified file (existing file will be overwritten!)',
    'quiet' :
    'do not print any messages to screen',
    'save_ebc' :
    'save a zero solution with applied EBCs (Dirichlet boundary conditions)',
    'save_ebc_nodes' :
    'save a zero solution with added non-zeros in EBC (Dirichlet boundary'
    ' conditions) nodes - scalar variables are shown using colors,'
    ' vector variables using arrows with non-zero components corresponding'
    ' to constrained components',
    'save_regions' :
    'save problem regions as meshes',
    'save_regions_as_groups' :
    'save problem regions in a single mesh but mark them by using different'
    ' element/node group numbers',
    'solve_not' :
    'do not solve (use in connection with --save-*)',
    'detect_band_gaps' :
    'detect frequency band gaps',
    'analyze_dispersion' :
    'analyze dispersion properties (low frequency domain)',
    'plot' :
    'plot frequency band gaps, assumes -b',
    'phase_velocity' :
    'compute phase velocity (frequency-independent mass only)',
    'list' :
    'list data, what can be one of: {terms, solvers}',
}

def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + sfepy.__version__)
    parser.add_argument('-a', '--app', action='store', dest='app',
                        choices=['bvp', 'homogen', 'bvp-mM', 'evp', 'phonon'],
                        default=None, help= helps['app'])
    parser.add_argument('--debug',
                        action='store_true', dest='debug',
                        default=False, help=helps['debug'])
    parser.add_argument('--debug-mpi',
                        action='store_true', dest='debug_mpi',
                        default=False, help=helps['debug_mpi'])
    parser.add_argument('-c', '--conf', metavar='"key : value, ..."',
                        action='store', dest='conf', type=str,
                        default=None, help= helps['conf'])
    parser.add_argument('-O', '--options', metavar='"key : value, ..."',
                        action='store', dest='app_options', type=str,
                        default=None, help=helps['options'])
    parser.add_argument('-d', '--define', metavar='"key : value, ..."',
                        action='store', dest='define_args', type=str,
                        default=None, help=helps['define'])
    parser.add_argument('-o', metavar='filename',
                        action='store', dest='output_filename_trunk',
                        default=None, help=helps['filename'])
    parser.add_argument('--format', metavar='format',
                        action='store', dest='output_format',
                        default=None, help=helps['output_format'])
    parser.add_argument('--save-restart', metavar='mode', type=int,
                        action='store', dest='save_restart',
                        default=None, help=helps['save_restart'])
    parser.add_argument('--load-restart', metavar='filename',
                        action='store', dest='load_restart',
                        default=None, help=helps['load_restart'])
    parser.add_argument('--log', metavar='file',
                        action='store', dest='log',
                        default=None, help=helps['log'])
    parser.add_argument('-q', '--quiet',
                        action='store_true', dest='quiet',
                        default=False, help=helps['quiet'])
    parser.add_argument('--save-ebc',
                        action='store_true', dest='save_ebc',
                        default=False, help=helps['save_ebc'])
    parser.add_argument('--save-ebc-nodes',
                        action='store_true', dest='save_ebc_nodes',
                        default=False, help=helps['save_ebc_nodes'])
    parser.add_argument('--save-regions',
                        action='store_true', dest='save_regions',
                        default=False, help=helps['save_regions'])
    parser.add_argument('--save-regions-as-groups',
                        action='store_true', dest='save_regions_as_groups',
                        default=False, help=helps['save_regions_as_groups'])
    parser.add_argument('--solve-not',
                        action='store_true', dest='solve_not',
                        default=False, help=helps['solve_not'])
    parser.add_argument('--phonon-band-gaps',
                        action='store_true', dest='detect_band_gaps',
                        default=False, help=helps['detect_band_gaps'])
    parser.add_argument('--phonon-dispersion',
                        action='store_true', dest='analyze_dispersion',
                        default=False, help=helps['analyze_dispersion'])
    parser.add_argument('--phonon-plot',
                        action='store_true', dest='plot',
                        default=False, help=helps['plot'])
    parser.add_argument('--phonon-phase-velocity',
                        action='store_true', dest='phase_velocity',
                        default=False, help=helps['phase_velocity'])
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--list', metavar='what',
                        action='store', dest='_list',
                        default=None, help=helps['list'])
    group.add_argument('filename_in', nargs='?')
    options, petsc_opts = parser.parse_known_args()

    if options._list is not None:
        if options._list == 'terms':
            print_terms()

        elif options._list == 'solvers':
            print_solvers()

        return

    if not (options.analyze_dispersion or options.detect_band_gaps):
            options.plot = False

    if options.debug:
        from sfepy.base.base import debug_on_error; debug_on_error()

    filename_in = options.filename_in
    output.set_output(filename=options.log,
                      quiet=options.quiet,
                      combined=options.log is not None)

    required, other = get_standard_keywords()
    required.remove('equations')
    if options.solve_not:
        required.remove('solver_[0-9]+|solvers')
        other.extend(['equations'])
    if options.detect_band_gaps and not options.analyze_dispersion:
        if 'solver_[0-9]+|solvers' in required:
            required.remove('solver_[0-9]+|solvers')
    if options.phase_velocity:
        required = [ii for ii in required if 'ebc' not in ii]

    conf = ProblemConf.from_file_and_options(filename_in, options,
                                             required, other,
                                             define_args=options.define_args)
    if conf.options.get('coefs') is None:
        if conf.get('equations') is None:
            ValueError('required missing: equations')

    app_mode = options.app
    if app_mode is None:
        if conf.options.get('coefs') is not None:
            if 'band_gaps' in conf.get(conf.options.coefs).keys():
                app_mode = 'phonon'

            else:
                app_mode = 'homogen'

        elif conf.options.get('evps') is not None:
            app_mode = 'evp'

        else:
            app_mode = 'bvp'

    opts = conf.options

    opts.save_restart = options.save_restart
    opts.load_restart = options.load_restart

    if app_mode == 'bvp':
        output_prefix = opts.get('output_prefix', 'sfepy:')
        app = PDESolverApp(conf, options, output_prefix)

    elif app_mode == 'homogen':
        from sfepy.homogenization.homogen_app import HomogenizationApp

        output_prefix = opts.get('output_prefix', 'homogen:')
        app = HomogenizationApp(conf, options, output_prefix)

    elif app_mode == 'bvp-mM':
        import sfepy.base.multiproc_mpi as multi_mpi

        if options.debug_mpi:
            multi_mpi.set_logging_level('debug')

        if multi_mpi.mpi_rank == multi_mpi.mpi_master:
            nslaves = multi_mpi.cpu_count() - 1
            opts.n_mpi_homog_slaves = nslaves
            output_prefix = opts.get('output_prefix', 'sfepy:')

            app = PDESolverApp(conf, options, output_prefix)
            if hasattr(opts, 'parametric_hook'):  # Parametric study.
                parametric_hook = conf.get_function(opts.parametric_hook)
                app.parametrize(parametric_hook)
            app()

            multi_mpi.master_send_task('finalize', None)
            return

        else:
            # MPI slave mode - calculate homogenized coefficients
            homogen_app = None
            done = False
            rank = multi_mpi.mpi_rank
            while not done:
                task, data = multi_mpi.slave_get_task('main slave loop')

                if task == 'init':  # data: micro_file, n_micro
                    output.set_output(filename='homog_app_mpi_%d.log' % rank,
                                      quiet=True)
                    micro_file, n_micro = data[:2]
                    required, other = get_standard_keywords()
                    required.remove('equations')
                    conf = ProblemConf.from_file(micro_file, required, other,
                                                 verbose=False)
                    options = Struct(output_filename_trunk=None)
                    homogen_app = HomogenizationApp(conf, options, 'micro:',
                                                    n_micro=n_micro)
                elif task == 'calculate':  # data: rel_def_grad, ts, iteration
                    macro_data, ts, iteration = data[:3]
                    homogen_app.setup_macro_data(macro_data)
                    homogen_app(ret_all=True, itime=ts.step, iiter=iteration)
                elif task == 'finalize':
                    done = True
            return

    elif app_mode == 'evp':
        output_prefix = opts.get('output_prefix', 'sfepy:')
        app = EVPSolverApp(conf, options, output_prefix)

    elif app_mode == 'phonon':
        from sfepy.homogenization.band_gaps_app import AcousticBandGapsApp

        output_prefix = opts.get('output_prefix', 'phonon:')
        app = AcousticBandGapsApp(conf, options, output_prefix)

    if hasattr(opts, 'parametric_hook'): # Parametric study.
        parametric_hook = conf.get_function(opts.parametric_hook)
        app.parametrize(parametric_hook)
    app()

if __name__ == '__main__':
    main()



import numpy as nm
import pytest

from sfepy.base.base import assert_
from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/2d/special/circle_in_square.mesh'

dim = 2

field_1 = {
    'name' : 'a_harmonic_field',
    'dtype' : 'real',
    'shape' : 'scalar',
    'region' : 'Omega',
    'approx_order' : 1,
}

variables = {
    't': ('unknown field', 'a_harmonic_field', 0),
    's': ('test field',    'a_harmonic_field', 't'),
}

regions = {
    'Omega' : 'all',
    'Gamma' : ('vertices of surface', 'facet'),
}

ebcs = {
    't_left' : ('Gamma', {'t.0' : 'ebc'}),
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

coef = 2.0
materials = {
    'coef' : ({'val' : coef},),
    'rhs' : 'rhs',
}

equations = {
    'Temperature' :
    """dw_laplace.i.Omega(coef.val, s, t)
       = - dw_volume_lvf.i.Omega(rhs.val, s)""",
}

solutions = {
    'sincos' : ('t', 'sin(3.0 * x) * cos(4.0 * y)',
                '-25.0 * %s * sin(3.0 * x) * cos(4.0 * y)' % coef),
    'poly' : ('t', '(x**2) + (y**2)', '4.0 * %s' % coef),
    'polysin' : ('t', '((x - 0.5)**3) * sin(5.0 * y)',
                 '%s * (6.0 * (x - 0.5) * sin(5.0 * y)'
                 ' - 25.0 * ((x - 0.5)**3) * sin(5.0 * y))' % coef),
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',

    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

solution = ['']
def ebc(ts, coor, **kwargs):
    expression = solution[0]
    val = tst.eval_coor_expression(expression, coor)
    return nm.atleast_1d(val)

def rhs(ts, coor, mode=None, expression=None, **kwargs):
    if mode == 'qp':
        if expression is None:
            expression = '0.0 * x'

        val = tst.eval_coor_expression(expression, coor)
        val.shape = (val.shape[0], 1, 1)

        return {'val' : val}

functions = {
    'ebc' : (ebc,),
    'rhs' : (rhs,),
}

@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    problem = Problem.from_conf(conf)
    return problem

def _build_rhs(sols):
    for sol in sols.values():
        assert_(len(sol) == 3)
    return sols

def test_msm_laplace(problem, output_dir):
    import os.path as op

    variables = problem.get_variables()
    materials = problem.get_materials()

    sols = _build_rhs(solutions)

    ok = True
    for sol_name, sol in sols.items():
        tst.report('testing', sol_name)
        var_name, sol_expr, rhs_expr=sol

        tst.report('sol:', sol_expr)
        tst.report('rhs:', rhs_expr)
        globals()['solution'][0] = sol_expr
        materials['rhs'].function.set_extra_args(expression=rhs_expr)
        problem.time_update()
        state = problem.solve(save_results=False)
        coor = variables[var_name].field.get_coor()
        ana_sol = tst.eval_coor_expression(sol_expr, coor)
        num_sol = state(var_name)

        ana_norm = nm.linalg.norm(ana_sol, nm.inf)
        ret = tst.compare_vectors(ana_sol, num_sol,
                                  allowed_error=ana_norm * 1e-2,
                                  label1='analytical %s' % var_name,
                                  label2='numerical %s' % var_name,
                                  norm=nm.inf)
        if not ret:
            tst.report('variable %s: failed' % var_name)

        fname = op.join(output_dir, 'test_msm_laplace_%s.vtk')
        out = {}
        astate = state.copy()
        astate.set_state(ana_sol)
        aux = astate.create_output()
        out['ana_t'] = aux['t']
        aux = state.create_output()
        out['num_t'] = aux['t']

        problem.domain.mesh.write(fname % sol_name, io='auto', out=out)

        ok = ok and ret

    assert_(ok)



import numpy as nm
import pytest

from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.solvers import Solver
import sfepy.base.testing as tst

# Mesh dimensions.
dims = nm.array([1, 1])

# Mesh resolution.
shape = [21, 21]

def mesh_hook(mesh, mode):
    """
    Generate the block mesh.
    """
    if mode == 'read':
        mesh = gen_block_mesh(dims, shape, [0, 0], name='user_block',
                              verbose=False)
        return mesh

    elif mode == 'write':
        pass

filename_mesh = UserMeshIO(mesh_hook)

regions = {
    'Omega' : 'all',
    'Gamma' : ('vertices of surface', 'facet'),
}

fields = {
    'temperature' : ('real', 1, 'Omega', 1),
}

variables = {
    't' : ('unknown field', 'temperature', 0),
    's' : ('test field',    'temperature', 't'),
}

ebcs = {
    't' : ('Gamma', {'t.0' : 0.0}),
}

integrals = {
    'i' : 2,
}

equations = {
    'eq' : 'dw_laplace.i.Omega(s, t) = 0'
}

solvers = {
    'evp0' : ('eig.scipy', {
        'method' : 'eig',
    }),
    'evp0h' : ('eig.scipy', {
        'method' : 'eigh',
    }),
    'evp0s' : ('eig.scipy', {
        'method' : 'eigs',
        'maxiter' : 100,
    }),
    'evp0sh' : ('eig.scipy', {
        'method' : 'eigsh',
        'maxiter' : 100,
    }),
    'evp1' : ('eig.sgscipy', {}),
    'evp2' : ('eig.scipy_lobpcg', {
        'i_max' : 100,
        'largest' : False,
    }),
    'evp4' : ('eig.slepc', {
        'method' : 'arnoldi',
        'problem' : 'hep',
        'i_max' : 100,
        'eps' : 1e-10,
        'which' : 'smallest_real',
    }),
    'evp5' : ('eig.matlab', {
        'maxit' : 100,
        'tol' : 1e-10,
        'which' : 'sr',
    }),
    'evp6' : ('eig.primme', {
        'maxiter' : 200,
        'tol' : 1e-10,
        'which' : 'sa',
    }),
}

eigs_expected = [nm.array([0.04904454, 0.12170685, 0.12170685,
                           0.19257998, 0.24082108]),
                 []]

can_fail = ['eig.slepc', 'eig.matlab', 'eig.primme']
can_miss = ['evp0'] # Depending on scipy version, evp0 can miss an
                    # eigenvalue.

@pytest.fixture(scope='module')
def data():
    import sys
    from sfepy.base.base import Struct
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])
    pb = Problem.from_conf(conf, init_solvers=False)
    pb.time_update()
    mtx = pb.equations.evaluate(mode='weak', dw_mode='matrix',
                                asm_obj=pb.mtx_a)

    return Struct(conf=conf, mtx=mtx)

def _list_eigenvalue_solvers(confs):
    d = []
    for key, val in confs.items():
        if val.kind.find('eig.') == 0:
            d.append(val)
    d.sort(key=lambda a: a.name)

    return d

def test_eigenvalue_solvers(data):
    from sfepy.base.base import IndexedStruct

    eig_confs = _list_eigenvalue_solvers(data.conf.solvers)

    all_n_eigs = [5, 0]

    ok = True
    tt = []
    for ii, n_eigs in enumerate(all_n_eigs):
        for eig_conf in eig_confs:
            tst.report(eig_conf.name)

            try:
                eig_solver = Solver.any_from_conf(eig_conf)

            except (ValueError, ImportError):
                if eig_conf.kind in can_fail:
                    continue

                else:
                    raise

            status = IndexedStruct()
            try:
                eigs, vecs = eig_solver(data.mtx, n_eigs=n_eigs,
                                        eigenvectors=True, status=status)

            except KeyboardInterrupt:
                raise

            except:
                if eig_conf.kind in can_fail:
                    continue

                else:
                    raise
            tt.append([' '.join((eig_conf.name, eig_conf.kind)),
                       status.time, n_eigs])

            tst.report(eigs)

            _ok = nm.allclose(eigs.real, eigs_expected[ii],
                              rtol=0.0, atol=1e-8)
            tt[-1].append(_ok)

            ok = ok and (_ok or (eig_conf.kind in can_fail)
                         or (eig_conf.name in can_miss))

    tt.sort(key=lambda x: x[1])
    tst.report('solution times:')
    for row in tt:
        tst.report('%.2f [s] : %s (%d) (ok: %s)'
                   % (row[1], row[0], row[2], row[3]))

    assert ok



"""
Proof-of-concept JAX-based terms supporting automatic differentiation.
"""
from functools import partial

import numpy as np

try:
    from jax import config
    config.update("jax_enable_x64", True)
    import jax
    import jax.numpy as jnp

except ImportError:
    print('install jax and jaxlib to use terms_jax.py!')
    raise

from sfepy.terms import Term
from sfepy.mechanics.matcoefs import lame_from_youngpoisson

@jax.jit
def get_strain(gu):
    return 0.5 * (gu + gu.T)

@jax.jit
def get_stress(lam, mu, gu):
    strain = get_strain(gu)
    eye = jnp.eye(gu.shape[-1])
    stress = lam * jnp.trace(strain) * eye + 2 * mu * strain
    return stress

@jax.jit
def ceval_elasticity_l(lam, mu, vbfg, ubfg, det, cu):
    gu = cu @ ubfg.transpose((0, 2, 1)) # This does DBD.
    stress = jax.vmap(get_stress, in_axes=[None, None, 0])(lam, mu, gu)
    vbfgd = vbfg * det
    val = jnp.sum(stress @ vbfgd, axis=0)
    return val

@partial(jax.jit, static_argnames=['plane'])
def ceval_elasticity_yp(young, poisson, plane, vbfg, ubfg, det, cu):
    gu = cu @ ubfg.transpose((0, 2, 1))
    lam, mu = lame_from_youngpoisson(young, poisson, plane=plane)
    stress = jax.vmap(get_stress, in_axes=[None, None, 0])(lam, mu, gu)
    vbfgd = vbfg * det
    val = jnp.sum(stress @ vbfgd, axis=0)
    return val

eval_elasticity_l = jax.jit(jax.vmap(ceval_elasticity_l,
                                     in_axes=[None, None, 0, 0, 0, 0]))
eval_jac_elasticity_l = jax.jit(jax.vmap(jax.jacobian(ceval_elasticity_l, -1),
                                         in_axes=[None, None, 0, 0, 0, 0]))
eval_lam_elasticity_l = jax.vmap(jax.jacobian(ceval_elasticity_l, 0),
                                 in_axes=[None, None, 0, 0, 0, 0])
eval_mu_elasticity_l = jax.vmap(jax.jacobian(ceval_elasticity_l, 1),
                                in_axes=[None, None, 0, 0, 0, 0])

eval_elasticity_yp = jax.jit(
    jax.vmap(ceval_elasticity_yp,
             in_axes=[None, None, None, 0, 0, 0, 0]),
    static_argnames=['plane'],
)
eval_jac_elasticity_yp = jax.jit(
    jax.vmap(jax.jacobian(ceval_elasticity_yp, -1),
             in_axes=[None, None, None, 0, 0, 0, 0]),
    static_argnames=['plane'],
)
eval_young_elasticity_yp = jax.vmap(jax.jacobian(ceval_elasticity_yp, 0),
                                    in_axes=[None, None, None, 0, 0, 0, 0])
eval_poisson_elasticity_yp = jax.vmap(jax.jacobian(ceval_elasticity_yp, 1),
                                      in_axes=[None, None, None, 0, 0, 0, 0])

class LinearElasticLADTerm(Term):
    r"""
    Homogeneous isotropic linear elasticity term differentiable w.r.t. material
    parameters :math:`\lambda`, :math:`\mu` (Lamé's parameters).

    :Definition:

    .. math::
        \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})\\ \mbox{ with } \\
        D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
        \lambda \ \delta_{ij} \delta_{kl}

    :Arguments:
        - material_1: :math:`\lambda` (Lamé's first parameter)
        - material_2: :math:`\mu` (Lamé's second parameter, shear modulus)
        - virtual/parameter_1: :math:`\ul{v}`
        - state/parameter_2: :math:`\ul{u}`
    """
    name = 'dw_lin_elastic_l_ad'
    arg_types = (('material_1', 'material_2', 'virtual', 'state'),)
    arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
                  'virtual' : ('D', 'state'), 'state' : 'D'}
    modes = ('weak',)
    diff_info = {'material_1' : 1, 'material_2' : 1}

    def get_fargs(self, material1, material2, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vgmap, _ = self.get_mapping(state)
        sgmap, _ = self.get_mapping(state)

        vecu = jnp.array(state().reshape((-1, vgmap.dim)))
        econn = state.field.get_econn(self.integration, self.region)
        # Transpose is required to have sfepy order (DBD).
        cu = vecu[econn].transpose((0, 2, 1))

        lam = material1[0, 0, 0, 0]
        mu = material2[0, 0, 0, 0]
        if diff_var is None:
            fun = eval_elasticity_l

        elif diff_var == state.name:
            fun = eval_jac_elasticity_l

        elif diff_var == 'material_1':
            fun = eval_lam_elasticity_l

        elif diff_var == 'material_2':
            fun = eval_mu_elasticity_l

        else:
            raise ValueError

        fargs = [lam, mu, vgmap.bfg, sgmap.bfg, vgmap.det, cu]

        return fun, fargs

    @staticmethod
    def function(out, fun, fargs):
        out[:] = np.asarray(fun(*fargs).reshape(out.shape))
        return 0

class LinearElasticYPADTerm(Term):
    r"""
    Homogeneous isotropic linear elasticity term differentiable w.r.t. material
    parameters :math:`E` (Young's modulus), :math:`\nu` (Poisson's ratio).

    :Definition:

    .. math::
        \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})\\ \mbox{ with } \\
        D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
        \lambda \ \delta_{ij} \delta_{kl}, \\ \mbox{ where } \\
        \lambda = E \nu / ((1 + \nu)(1 - 2\nu)), \\ \mu = E / 2(1 + \nu)

    :Arguments:
        - material_1: :math:`E`
        - material_2: :math:`\nu`
        - virtual/parameter_1: :math:`\ul{v}`
        - state/parameter_2: :math:`\ul{u}`
    """
    name = 'dw_lin_elastic_yp_ad'
    arg_types = (('material_1', 'material_2', 'virtual', 'state'),)
    arg_shapes = {'material_1' : '1, 1', 'material_2' : '1, 1',
                  'virtual' : ('D', 'state'), 'state' : 'D'}
    modes = ('weak',)
    diff_info = {'material_1' : 1, 'material_2' : 1}

    def get_fargs(self, material1, material2, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vgmap, _ = self.get_mapping(state)
        sgmap, _ = self.get_mapping(state)

        vecu = jnp.array(state().reshape((-1, vgmap.dim)))
        econn = state.field.get_econn(self.integration, self.region)
        # Transpose is required to have sfepy order (DBD).
        cu = vecu[econn].transpose((0, 2, 1))

        young = material1[0, 0, 0, 0]
        poisson = material2[0, 0, 0, 0]
        if diff_var is None:
            fun = eval_elasticity_yp

        elif diff_var == state.name:
            fun = eval_jac_elasticity_yp

        elif diff_var == 'material_1':
            fun = eval_young_elasticity_yp

        elif diff_var == 'material_2':
            fun = eval_poisson_elasticity_yp

        else:
            raise ValueError

        fargs = [young, poisson, 'strain', vgmap.bfg, sgmap.bfg, vgmap.det, cu]

        return fun, fargs

    @staticmethod
    def function(out, fun, fargs):
        out[:] = np.asarray(fun(*fargs).reshape(out.shape))
        return 0

@jax.jit
def ceval_mass(density, vbf, ubf, det, cu):
    ru = density * cu @ ubf.transpose((0, 2, 1))
    vbfd = vbf * det
    val = jnp.sum(ru @ vbfd, axis=0)
    return val

eval_mass = jax.jit(jax.vmap(ceval_mass,
                             in_axes=[None, 0, 0, 0, 0]))
eval_jac_mass = jax.jit(jax.vmap(jax.jacobian(ceval_mass, -1),
                                 in_axes=[None, 0, 0, 0, 0]))
eval_density_mass = jax.vmap(jax.jacobian(ceval_mass, 0),
                             in_axes=[None, 0, 0, 0, 0])

class MassADTerm(Term):
    r"""
    Homogeneous mass term differentiable w.r.t. the material parameter.

    :Definition:

    .. math::
        \int_{\cal{D}} \rho \ul{v} \cdot \ul{u}

    :Arguments:
        - material_1: :math:`\rho`
        - virtual: :math:`\ul{v}`
        - state: :math:`\ul{u}`
    """
    name = 'dw_mass_ad'
    arg_types = (('material', 'virtual', 'state'),)
    arg_shapes = {'material' : '1, 1',
                  'virtual' : ('D', 'state'), 'state' : 'D'}
    modes = ('weak',)
    diff_info = {'material' : 1}

    def get_fargs(self, material_density, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vgmap, _ = self.get_mapping(state)
        sgmap, _ = self.get_mapping(state)

        vecu = jnp.array(state().reshape((-1, vgmap.dim)))
        econn = state.field.get_econn(self.integration, self.region)
        # Transpose is required to have sfepy order (DBD).
        cu = vecu[econn].transpose((0, 2, 1))

        density = material_density[0, 0, 0, 0]
        if diff_var is None:
            fun = eval_mass

        elif diff_var == state.name:
            fun = eval_jac_mass

        elif diff_var == 'material':
            fun = eval_density_mass

        else:
            raise ValueError

        bs = partial(np.broadcast_to, shape=(cu.shape[0],) + vgmap.bf.shape[1:])
        fargs = [density, bs(vgmap.bf), bs(sgmap.bf), vgmap.det, cu]

        return fun, fargs

    @staticmethod
    def function(out, fun, fargs):
        out[:] = np.asarray(fun(*fargs).reshape(out.shape))
        return 0

def _get_Eq(gu):
    return 0.5 * (gu + gu.T + gu.T @ gu)

def _get_E(bfg, cu):
    gu = cu @ bfg.transpose((0, 2, 1)) # This does DBD.
    val = jax.vmap(_get_Eq, in_axes=[0])(gu)
    return val

_get_dEv = jax.jacobian(_get_E, -1)

def get_neohook_strain_energy(mu, C):
    # (*) 0.5 *
    dim = C.shape[-1]
    trc  = jnp.trace(C)
    if dim == 2:
        trc += 1.0 # Plane strain.

    W = mu * (jnp.linalg.det(C)**(-1.0/3.0) * trc - 3.0)
    return W

_get_neohook_stress_2pk = jax.grad(get_neohook_strain_energy, -1)

def get_neohook_stress_2pk(mu, gu):
    eye = np.eye(gu.shape[-1])
    F = gu + eye
    # (*) 2 *
    S = _get_neohook_stress_2pk(mu, F.T @ F)
    return S

# This is very slow, worsens with problem size w.r.t. dw_tl_neohook and order 2
# is out of memory.
# @jax.jit
def ceval_neohook0(mu, vbfg, ubfg, det, cu):
    gu = cu @ ubfg.transpose((0, 2, 1)) # This does DBD.
    stress = jax.vmap(get_neohook_stress_2pk, in_axes=[None, 0])(mu, gu)
    dEv = _get_dEv(vbfg, cu) * det[..., None, None]
    val = jnp.sum(stress[..., None, None] * dEv, axis=[0, 1, 2])
    return val

def get_neohook_strain_energy_f(mu, F):
    dim = F.shape[-1]
    trf = jnp.trace(F.T @ F)
    if dim == 2:
        trf += 1.0 # Plane strain

    W = 0.5 * mu * (jnp.linalg.det(F)**(-2.0/3.0) * trf - 3.0)
    return W

_get_neohook_stress_1pk = jax.grad(get_neohook_strain_energy_f, -1)

def get_neohook_stress_1pk(mu, gu):
    eye = np.eye(gu.shape[-1])
    F = gu + eye
    S = _get_neohook_stress_1pk(mu, F)
    return S

# This is slow, improves with problem size w.r.t. sfepy and order 2 works, with
# refine about 5x slower than sfepy.
@jax.jit
def ceval_neohook(mu, vbfg, ubfg, det, cu):
    gu = cu @ ubfg.transpose((0, 2, 1)) # This does DBD.
    stress = jax.vmap(get_neohook_stress_1pk, in_axes=[None, 0])(mu, gu)
    vbfgd = vbfg * det
    val = jnp.sum(stress @ vbfgd, axis=0)
    return val

eval_neohook = jax.jit(jax.vmap(ceval_neohook,
                                in_axes=[None, 0, 0, 0, 0]))
eval_jac_neohook = jax.jit(jax.vmap(jax.jacobian(ceval_neohook, -1),
                                    in_axes=[None, 0, 0, 0, 0]))
eval_mu_neohook = jax.jit(jax.vmap(jax.jacobian(ceval_neohook, 0),
                                   in_axes=[None, 0, 0, 0, 0]))

class NeoHookeanTLADTerm(Term):
    r"""
    Homogeneous Hyperelastic neo-Hookean term differentiable w.r.t. the
    material parameter. Effective stress :math:`S_{ij} = \mu
    J^{-\frac{2}{3}}(\delta_{ij} - \frac{1}{3}C_{kk}C_{ij}^{-1})`.

    :Definition:

    .. math::
        \int_{\Omega} S_{ij}(\ul{u}) \delta E_{ij}(\ul{u};\ul{v})

    :Arguments:
        - material : :math:`\mu`
        - virtual  : :math:`\ul{v}`
        - state    : :math:`\ul{u}`
    """
    name = 'dw_tl_he_neohook_ad'
    arg_types = ('material', 'virtual', 'state')
    arg_shapes = {'material' : '1, 1', 'virtual' : ('D', 'state'),
                  'state' : 'D'}
    modes = ('weak',)
    diff_info = {'material' : 1}
    geometries = ['2_3', '2_4', '3_4', '3_8']

    def get_fargs(self, material, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vgmap, _ = self.get_mapping(virtual)
        sgmap, _ = self.get_mapping(state)

        vecu = state().reshape((-1, vgmap.dim))
        econn = state.field.get_econn(self.integration, self.region)
        # Transpose is required to have sfepy order (DBD).
        cu = vecu[econn].transpose((0, 2, 1))

        mu = material[0, 0, 0, 0]
        if diff_var is None:
            fun = eval_neohook

        elif diff_var == state.name:
            fun = eval_jac_neohook

        elif diff_var == 'material':
            fun = eval_mu_neohook

        else:
            raise ValueError

        fargs = [mu, vgmap.bfg, sgmap.bfg, vgmap.det, cu]
        fargs = [jax.device_put(val) for val in fargs]
        return fun, fargs

    @staticmethod
    def function(out, fun, fargs):
        out[:] = np.asarray(fun(*fargs).reshape(out.shape))
        return 0

def get_ogden_strain_energy_f(mu, alpha, F):
    J = jnp.linalg.det(F)
    C = J**(-2.0/3.0) * F.T @ F # Isochoric C.
    lams2, Ns = jnp.linalg.eigh(C)
    a5 = 0.5 * alpha
    W = (mu / alpha) * (jnp.sum(lams2**a5) - 3.0)
    return W

_get_ogden_stress_1pk = jax.grad(get_ogden_strain_energy_f, -1)

def get_ogden_stress_1pk(mu, alpha, gu):
    eye = np.eye(gu.shape[-1])
    F = gu + eye
    S = _get_ogden_stress_1pk(mu, alpha, F)
    return S

@jax.jit
def ceval_ogden(mu, alpha, vbfg, ubfg, det, cu):
    gu = cu @ ubfg.transpose((0, 2, 1)) # This does DBD.
    stress = jax.vmap(get_ogden_stress_1pk,
                      in_axes=[None, None, 0])(mu, alpha, gu)
    vbfgd = vbfg * det
    val = jnp.sum(stress @ vbfgd, axis=0)
    return val

eval_ogden = (jax.vmap(ceval_ogden,
                              in_axes=[None, None, 0, 0, 0, 0]))
eval_jac_ogden = (jax.vmap(jax.jacobian(ceval_ogden, -1),
                                  in_axes=[None, None, 0, 0, 0, 0]))
eval_mu_ogden = jax.jit(jax.vmap(jax.jacobian(ceval_ogden, 0),
                                 in_axes=[None, None, 0, 0, 0, 0]))
eval_alpha_ogden = jax.jit(jax.vmap(jax.jacobian(ceval_ogden, 1),
                                    in_axes=[None, None, 0, 0, 0, 0]))

class OgdenTLADTerm(Term):
    r"""
    Homogeneous hyperelastic Ogden model term differentiable w.r.t. the
    material parameters, with the strain energy density

    .. math::
        W = \frac{\mu}{\alpha} \, \left(
            \bar\lambda_1^{\alpha} + \bar\lambda_2^{\alpha}
             + \bar\lambda_3^{\alpha} - 3 \right) \; ,

    where :math:`\lambda_k, k=1, 2, 3` are the principal stretches, whose
    squares are the principal values of the right Cauchy-Green deformation
    tensor :math:`\mathbf{C}`. For more details see :class:`OgdenTLTerm
    <sfepy.terms.terms_hyperelastic_tl.OgdenTLTerm>`.

    WARNING: The current implementation fails to compute the tangent matrix
    when :math:`\mathbf{C}` has multiple eigenvalues (e.g. zero deformation).
    In that case nans are returned, as a result of dividing by zero. See [1],
    Section 11.2.3, page 385.

    [1] Borst, R. de, Crisfield, M.A., Remmers, J.J.C., Verhoosel, C.V., 2012.
    Nonlinear Finite Element Analysis of Solids and Structures, 2nd edition.
    ed. Wiley, Hoboken, NJ.

    :Definition:

    .. math::
        \int_{\Omega} S_{ij}(\ul{u}) \delta E_{ij}(\ul{u};\ul{v})

    :Arguments:
        - material_1 : :math:`\mu`
        - material_2 : :math:`\alpha`
        - virtual    : :math:`\ul{v}`
        - state      : :math:`\ul{u}`
    """
    name = 'dw_tl_he_ogden_ad'
    arg_types = ('material_mu', 'material_alpha', 'virtual', 'state')
    arg_shapes = {'material_mu' : '1, 1', 'material_alpha' : '1, 1',
                  'virtual' : ('D', 'state'), 'state' : 'D'}
    modes = ('weak',)
    diff_info = {'material_mu' : 1, 'material_alpha' : 1}
    geometries = ['3_4', '3_8']

    def get_fargs(self, material_mu, material_alpha, virtual, state,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vgmap, _ = self.get_mapping(virtual)
        sgmap, _ = self.get_mapping(state)

        vecu = state().reshape((-1, vgmap.dim))
        econn = state.field.get_econn(self.integration, self.region)
        # Transpose is required to have sfepy order (DBD).
        cu = vecu[econn].transpose((0, 2, 1))

        mu = material_mu[0, 0, 0, 0]
        alpha = material_alpha[0, 0, 0, 0]

        if diff_var is None:
            fun = eval_ogden

        elif diff_var == state.name:
            fun = eval_jac_ogden

        elif diff_var == 'material_mu':
            fun = eval_mu_ogden

        elif diff_var == 'material_alpha':
            fun = eval_alpha_ogden

        else:
            raise ValueError

        fargs = [mu, alpha, vgmap.bfg, sgmap.bfg, vgmap.det, cu]
        fargs = [jax.device_put(val) for val in fargs]
        return fun, fargs

    @staticmethod
    def function(out, fun, fargs):
        out[:] = np.asarray(fun(*fargs).reshape(out.shape))
        return 0



import os.path as op
import numpy as nm
import scipy.sparse as sp

from sfepy.base.base import assert_
import sfepy.base.testing as tst

def test_sparse_matrix_hdf5(output_dir):
    from sfepy.base.ioutils import (write_sparse_matrix_hdf5,
                                    read_sparse_matrix_hdf5, pt)
    if pt is None:
        tst.report('skipped (no pytables)')
        return

    filename = op.join(output_dir, 'mtx.h5')

    aux = nm.random.rand(5, 5)
    aux[1,:] = aux[:,2] = aux[3,:] = 0.0

    mtx = sp.csr_matrix(aux, dtype = nm.float64)
    tst.report('saving matrix into %s...' % filename)
    write_sparse_matrix_hdf5(filename, mtx)
    tst.report('reading...')
    mtx2 = read_sparse_matrix_hdf5(filename)
    tst.report('difference:\n%s' % (mtx2 - mtx).__repr__())

    assert_(mtx.shape == mtx2.shape)
    assert_(mtx.dtype == mtx2.dtype)
    assert_(mtx.format == mtx2.format)
    assert_(nm.allclose(mtx.data, mtx2.data))
    assert_(nm.allclose(mtx.indices, mtx2.indices))
    assert_(nm.allclose(mtx.indptr, mtx2.indptr))

def test_recursive_dict_hdf5(output_dir):
    from sfepy.base.ioutils import write_dict_hdf5, read_dict_hdf5, pt
    if pt is None:
        tst.report('skipped (no pytables)')
        return

    filename = op.join(output_dir, 'dict.h5')

    test = {'A' : 0, 'B' : {'C' : [0, 1],
                            'D' : {'E' : {'F' : {'G' : 2.0}}}}}

    tst.report('%s' % test)
    tst.report('saving into %s...' % filename)
    write_dict_hdf5(filename, test)
    tst.report('reading...')
    test2 = read_dict_hdf5(filename)
    tst.report('%s' % test2)

    assert_(test == test2)



r"""
The linear elastodynamics solution of an iron plate impact problem with
identification of material parameters from simulated measurement data.

Find :math:`\ul{u}` such that:

.. math::
    \int_{\Omega} \rho \ul{v} \pddiff{\ul{u}}{t}
    + \int_{\Omega} D_{ijkl}\ e_{ij}(\ul{v}) e_{kl}(\ul{u})
    = 0
    \;, \quad \forall \ul{v} \;,

where

.. math::
    D_{ijkl} = \mu (\delta_{ik} \delta_{jl}+\delta_{il} \delta_{jk}) +
    \lambda \ \delta_{ij} \delta_{kl} \;,

    \lambda = E \nu / ((1 + \nu)(1 - 2\nu)), \\ \mu = E / 2(1 + \nu)
    \;.

Usage Examples
--------------

- Run without the identification::

    sfepy-run sfepy/examples/linear_elasticity/elastodynamic_identification.py
    sfepy-view output/edi/user_block.h5 -f u:wu:f1e3:p0 1:vw:p0

- Get help::

    python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py -h

- Run the identification with default parameters, show live plot of
  convergence and launch ipython shell after the computation::

    python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py --plot-log --shell

  Result figures are in output/edi, if not changed using --output-dir option.

- Check the Jacobian matrix by finite differences::

    python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py --opt-conf=max_nfev=1 --check-jac --shell

- Identify also the damping parameters (zero by default)::

    python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py --par-names=young,poisson,density,alpha,beta --plot-log --shell

See also :ref:`linear_elasticity-elastodynamic`.
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import sys
import os.path as op
from functools import partial

import numpy as nm
import scipy.sparse as sps
import matplotlib.pyplot as plt

from sfepy.base.base import output, Struct, IndexedStruct
from sfepy.base.conf import dict_from_string as parse_as_dict
from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh
from sfepy.homogenization.utils import define_box_regions
from sfepy.solvers import register_solver, Solver
from sfepy.base.timing import Timer
import sfepy.mechanics.matcoefs as mc

def define(
        young=200e9, poisson=0.3, density=7800,
        alpha=0.0, beta=0.0,
        plane='strain',
        dims=(1e-2, 2.5e-3, 2.5e-3),
        shape=(21, 6, 6),
        v0=1.0,
        ct1=1.5,
        dt=None,
        active_only=False,
        save_times=20,
        output_dir='output/edi',
        **kwargs,
):
    """
    Parameters
    ----------
    young, poisson, density: material parameters
    plane: plane strain or stress hypothesis
    dims: physical dimensions of the block (L, d, x)
    shape: numbers of mesh vertices along each axis
    v0: initial impact velocity
    ct1: final time in L / "longitudinal wave speed" units
    dt: time step (None means automatic)
    save_times: number of time steps to save
    output_dir: output directory
    """
    dim = len(dims)

    lam, mu = mc.lame_from_youngpoisson(young, poisson, plane=plane)
    # Longitudinal and shear wave propagation speeds.
    cl = nm.sqrt((lam + 2.0 * mu) / density)
    cs = nm.sqrt(mu / density)

    # Element size.
    L, d = dims[:2]
    H = L / (nm.max(shape) - 1)

    # Time-stepping parameters.
    if dt is None:
        # For implicit schemes, dt based on the Courant number C0 = dt * cl / H
        # equal to 1.
        dt = H / cl # C0 = 1

    t1 = ct1 * L / cl

    def mesh_hook(mesh, mode):
        """
        Generate the block mesh.
        """
        if mode == 'read':
            mesh = gen_block_mesh(dims, shape, 0.5 * nm.array(dims),
                                  name='user_block', verbose=False)
            return mesh

        elif mode == 'write':
            pass

    filename_mesh = UserMeshIO(mesh_hook)

    bbox = [[0] * dim, dims]
    regions = define_box_regions(dim, bbox[0], bbox[1], 1e-5)

    dx = dims[0] / (shape[0] - 1)
    sx = int(0.2 * (shape[0] - 1)) * dx
    regions.update({
        'Omega' : 'all',
        'Plane' : ('vertices in (x > %.12f) & (x < %.12f)'
                   % (0.99999 * sx, 1.00001 * sx),
                   'facet'),
    })
    if dim == 3:
        regions.update({
            'Sensor' : ('r.Plane *v r.Top *v r.Far', 'vertex'),
        })

    else:
        regions.update({
            'Sensor' : ('r.Plane *v r.Top', 'vertex'),
        })

    # Iron.
    materials = {
        'm' : ({
            'young' : young,
            'poisson' : poisson,
            'density': density,
        },),
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', 1),
    }

    integrals = {
        'i' : 2,
    }

    # Notes:
    # 1. The order of the variables in the solution vector is specified here
    #    (3rd tuple member), since that specific order is expected by the
    #    elastodynamic time-stepping solvers.
    # 2. For the same reason, we won't explicitly define below the equations
    #    du = du/dt and ddu = ddu/dt - these are implicitly defined by
    #    the time-stepping solver. see the `step()` method of the solvers.
    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'du' : ('unknown field', 'displacement', 1),
        'ddu' : ('unknown field', 'displacement', 2),
        'v' : ('test field', 'displacement', 'u'),
        'dv' : ('test field', 'displacement', 'du'),
        'ddv' : ('test field', 'displacement', 'ddu'),
    }
    # The mapping of variables for the elastodynamics solvers - keys are given,
    # values correspond to the names of the actual variables.
    var_names = {'u' : 'u', 'du' : 'du', 'ddu' : 'ddu'}

    ebcs = {
        'Impact' : ('Left', {'u.0' : 0.0, 'du.0' : 0.0, 'ddu.0' : 0.0}),
    }
    if dim == 3:
        ebcs.update({
            'Symmtery-y' : ('Near',
                            {'u.1' : 0.0, 'du.1' : 0.0, 'ddu.1' : 0.0}),
            'Symmetry-z' : ('Bottom',
                            {'u.2' : 0.0, 'du.2' : 0.0, 'ddu.2' : 0.0}),
        })

    def get_ic(coor, ic, mode='u'):
        val = nm.zeros_like(coor)
        if mode == 'u':
            val[:, 0] = 0.0

        elif mode == 'du':
            val[:, 0] = -1.0

        return val

    functions = {
        'get_ic_u' : (get_ic,),
        'get_ic_du' : (lambda coor, ic: get_ic(coor, None, mode='du'),),
    }

    ics = {
        'ic' : ('Omega', {'u.all' : 'get_ic_u', 'du.all' : 'get_ic_du'}),
    }

    equations = {
        'eq' :
        f"""
         + dw_mass_ad.i.Omega(m.density, ddv, ddu)
         + {alpha} * dw_mass_ad.i.Omega(m.density, dv, du)
         + {beta} * dw_lin_elastic_yp_ad.i.Omega(m.young, m.poisson, dv, du)
         + dw_lin_elastic_yp_ad.i.Omega(m.young, m.poisson, v, u) = 0""",
    }

    solvers = {
        'lsd' : ('ls.auto_direct', {
        # 'lsd' : ('ls.scipy_direct', {
            # Reuse the factorized linear system from the first time step.
            'use_presolve' : True,
            # Speed up the above by omitting the matrix digest check used
            # normally for verification that the current matrix corresponds to
            # the factorized matrix stored in the solver instance. Use with
            # care!
            'use_mtx_digest' : False,
            # Increase when getting MUMPS error -9.
            'memory_relaxation' : 50,
        }),
        'newton' : ('nls.newton', {
            'i_max'      : 1,
            'eps_a'      : 1e-6,
            'eps_r'      : 1e-6,
        }),
        'tsn' : ('ts.newmark', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'beta' : 0.25,
            'gamma' : 0.5,

            'var_names' : var_names,
            'verbose' : 1,
        }),
    }

    options = {
        'ts' : 'tsn',
        'tsc' : None,
        'nls' : 'newton',
        'ls' : 'lsd',

        'save_times' : save_times,

        'active_only' : active_only,
        'auto_transform_equations' : False,

        'output_format' : 'h5',
        'output_dir' : output_dir,
    }

    return locals()

from sfepy.solvers.ts_solvers import NewmarkTS, _cache

class NewmarkSATS(NewmarkTS):
    r"""
    Solve elastodynamics problems by the Newmark method.

    The method was introduced in [1]. Common settings [2]:

    ==================== ======== ==== ===== ==========
    name                 kind     beta gamma Omega_crit
    ==================== ======== ==== ===== ==========
    trapezoidal rule:    implicit 1/4  1/2   unconditional
    linear acceleration: implicit 1/6  1/2   :math:`2\sqrt{3}`
    Fox-Goodwin:         implicit 1/12 1/2   :math:`\sqrt{6}`
    central difference:  explicit 0    1/2   2
    ==================== ======== ==== ===== ==========

    All of these methods are 2-order of accuracy.

    [1] Newmark, N. M. (1959) A method of computation for structural dynamics.
    Journal of Engineering Mechanics, ASCE, 85 (EM3) 67-94.

    [2] Arnaud Delaplace, David Ryckelynck: Solvers for Computational Mechanics
    """
    name = 'ts.newmark_sa'

    def create_nlst(self, nls, dt, gamma, beta, u0, e0, v0, a0, pack, unpack):
        dt2 = dt**2
        iue, iu, ie, iv, ia = pack.indices

        if iue != iu:
            raise ValueError('extra variables not supported!')

        cc0 = (1.0 - gamma) * dt
        cc = gamma * dt
        ck0 = (0.5 - beta) * dt2
        ck = beta * dt2

        def v(a):
            return v0 + cc0 * a0 + cc * a

        def u(a):
            return u0 + dt * v0 + ck0 * a0 + ck * a

        def fun(at):
            vec = nm.r_[u(at), v(at), at]

            aux = nls.fun(vec)

            rt = aux[iu] + aux[iv] + aux[ia]
            return rt

        @_cache(self, 'matrix', self.conf.is_linear)
        def fun_grad(at):
            vec = None if self.conf.is_linear else nm.r_[u(at), v(at), at]
            M, C, K = self.get_matrices(nls, vec, unpack)

            Kt = M + cc * C + ck * K
            return Kt

        @_cache(self, 'matrix_sa', self.conf.is_linear and (self.ts.step > 0))
        def efun_grad(at):
            vec = None if self.conf.is_linear else nm.r_[u(at), v(at), at]
            M, C, K = self.get_matrices(nls, vec, unpack)

            size = at.shape[0]
            I = sps.eye(size, dtype=nm.float64)
            zz = sps.dia_matrix((size, size), dtype=nm.float64)
            if self.ts.step > 0:
                duda = -ck * I
                dvda = -cc * I

            else:
                duda = dvda = zz
                # The initial state might depend on parameters - it does not in
                # this example as u0 = 0, v0 = const and a0 = 0, but the
                # following code should work in the general case if the next
                # line is commented out.
                K = C = zz

            return sps.bmat([[K, C, M],
                             [I, zz, duda],
                             [zz, I, dvda]])

        @_cache(self, 'matrix_sa0', self.conf.is_linear and (self.ts.step > 0))
        def efun_grad0(at):
            if self.ts.step > 0:
                size = at.shape[0]
                I = sps.eye(size, dtype=nm.float64)
                zz = sps.dia_matrix((size, size), dtype=nm.float64)
                duda0 = -ck0 * I
                dudv0 = -dt * I
                dudu0 = -I
                dvda0 = -cc0 * I
                dvdv0 = -I
                return sps.bmat([[zz, zz, zz],
                                 [dudu0, dudv0, duda0],
                                 [zz, dvdv0, dvda0]])

            else:
                return sps.csr_matrix((self.sa_info.n_dof, self.sa_info.n_dof),
                                      dtype=nm.float64)

        def efun_grad_par(at):
            mtx = nm.zeros((self.sa_info.n_dof, self.sa_info.n_par),
                           dtype=nm.float64, order='F')
            mtx = self.context.equations.evaluate(
                mode='weak', dw_mode='sensitivity',
                diff_vars=self.sa_info.par_names, asm_obj=mtx,
            )
            for ic, name in enumerate(self.sa_info.par_names):
                if name == 'alpha': # = M v
                    info = self.sa_info.par_info['alpha']

                elif name == 'beta': # = K v
                    info = self.sa_info.par_info['beta']

                else:
                    continue

                term = self.context.equations[info[1]].terms[info[2]]
                val, iels, status = term.evaluate(mode='weak',
                                                  diff_var=None,
                                                  standalone=False,
                                                  ret_status=True)
                val /= term.sign
                term.assemble_to(mtx[:, ic], val, iels)

            Vt = mtx[iu] + mtx[iv] + mtx[ia]
            zz = nm.zeros_like(Vt, dtype=nm.float64)
            return nm.block([[Vt],
                             [zz],
                             [zz]])

        nlst = nls.copy()
        nlst.fun = fun
        nlst.fun_grad = fun_grad
        nlst.efun_grad = efun_grad
        nlst.efun_grad0 = efun_grad0
        nlst.efun_grad_par = efun_grad_par
        nlst.u = u
        nlst.v = v

        self.nlst = nlst

        return nlst

register_solver(NewmarkSATS)

def apply_sensor(pb, ts, state, out):
    us = state['u'].get_state_in_region(pb.domain.regions['Sensor'])
    output('sensor', ts.step, us)
    out.append(us)

def update_pars(materials, equations, pars, par_names, par_info):
    """
    Materials and options are updated in place.
    """
    done = set()
    for key in par_names:
        if key in done: continue

        info = par_info[key]
        ip = par_names.index(key)
        if info[0] == 'term': # Update term coefficient.
            equations[info[1]].terms[info[2]].sign = pars[ip]

        else: # Update materials.
            mat = materials[info[0]]
            mkeys = mat.get_keys(info[1])
            for mkey in mkeys:
                if key == info[2]:
                    val = pars[ip]

                else:
                    raise ValueError

                mat.datas[mkey][info[2]][:, :] = val

    output('updated:', par_names)

def eval_fun(pars, data, pb, options, par_names, par_info, opt_data, plog,
             inodir):
    opt_data.tfun.start()

    materials = pb.get_materials()
    update_pars(materials, pb.equations, pars, par_names, par_info)

    pb.ts.set_step() # Reset ts.
    pb.get_solver().clear_lin_solver() # No digest -> clear manually.

    out = []
    pb.solve(save_results=False, step_hook=partial(apply_sensor, out=out))
    usens = nm.concatenate(out)

    sim_data = usens[:, -1]
    res = sim_data - data

    plog(*(tuple(pars) + (nm.linalg.norm(res),)))

    times = pb.ts.times
    fig, ax = plt.subplots()
    ax.plot(times, sim_data, label='sim')
    ax.plot(times, data, ls='', marker='o', mew=0, label='exp')
    ax.plot(times, opt_data.true_data, ls=':', label='true')
    ax.legend()
    plt.tight_layout()
    fig.savefig(inodir(f'res{opt_data.n_fun:05d}.png'), bbox_inches='tight')

    opt_data.n_fun += 1
    opt_data.usens = usens

    dt = opt_data.tfun.stop()
    output(f'function evaluation done in {dt} s.')

    return res

def eval_jac_fd(pars, *args, **kwargs):
    v0 = eval_fun(pars, *args, **kwargs)

    eps = 1e-6
    jac = []
    for ip, par in enumerate(pars):
        pars1 = pars.copy()
        dp = eps * par
        pars1[ip] += dp

        v1 = eval_fun(pars1, *args, **kwargs)
        jac.append((v1 - v0) / dp)

    jac = nm.array(jac).T

    return jac

def eval_jac(pars, data, pb, options, par_names, par_info, opt_data, plog,
             inodir):
    opt_data.tjac.start()

    materials = pb.get_materials()
    update_pars(materials, pb.equations, pars, par_names, par_info)

    pb.ts.set_step() # Reset ts.
    pb.get_solver().clear_lin_solver() # No digest -> clear manually.

    variables = pb.get_initial_state()
    n_dof = variables.adi.n_dof_total
    n_par = len(par_names)

    _tss = pb.get_solver()
    conf = _tss.conf.copy(name='tss_sa')
    conf.kind = 'ts.newmark_sa'

    tss = Solver.any_from_conf(conf, nls=_tss.nls, tsc=_tss.tsc, context=pb)
    tss.sa_info = Struct(
        par_info=par_info, par_names=par_names, n_par=n_par, n_dof=n_dof,
    )
    tss.matrix_sa = tss.matrix_sa0 = tss.matrix_pars = None

    ls = Solver.any_from_conf(tss.nls.lin_solver.conf)

    _init_fun, prestep_fun, _poststep_fun = pb.get_tss_functions(
        update_bcs=True, update_materials=True,
        save_results=False,
        step_hook=None, post_process_hook=None,
    )

    fu = pb.fields['displacement']
    ii = fu.get_dofs_in_region(pb.domain.regions['Sensor'])[0]
    idof = int(fu.n_components * ii + 2 + variables.adi.indx['u'].start)

    def init_fun(tss, ts, vec0):
        vec0 = _init_fun(ts, vec0)
        # Previous step dy/da.
        tss.gy0 = nm.zeros((tss.sa_info.n_dof, tss.sa_info.n_par),
                           dtype=nm.float64, order='F')
        return vec0

    jac = []
    def poststep_fun(tss, ts, vec):
        # EBCs need to be applied here because of algebraically computed
        # variables in elastodynamics solvers.
        variables.set_state(vec, pb.active_only, apply_ebc=True)

        unpack = tss.unpack
        utp, etp, vtp, atp = unpack(vec)
        if ts.step == 0:
            nlst = tss.create_nlst(tss.nls, ts.dt, tss.conf.gamma, tss.conf.beta,
                                   utp, etp, vtp, atp, tss.pack, unpack)

        else:
            nlst = tss.nlst

        Atp = nlst.efun_grad(atp)
        Wtp = nlst.efun_grad0(atp)
        Vtp = nlst.efun_grad_par(atp)

        ry = -(Vtp + Wtp @ tss.gy0)

        if options.multi_rhs:
            gy = ls(ry, mtx=Atp)

        else:
            gy = nm.empty_like(ry)
            for ic in range(gy.shape[1]):
                gy[:, ic] = ls(ry[:, ic], mtx=Atp)

        if ts.step == 0:
            # No digest -> clear manually.
            ls.clear()

        jac.append(gy[idof])

        tss.gy0 = gy
        pb.advance(ts)

        return vec

    tss.set_dof_info(variables.adi)
    status = IndexedStruct()
    tss(variables.get_state(pb.active_only, force=True),
        init_fun=init_fun,
        prestep_fun=prestep_fun,
        poststep_fun=poststep_fun,
        status=status)

    jac = nm.array(jac)

    if options.check_jac:
        jac_fd = eval_jac_fd(pars, data, pb, options, par_names, par_info,
                             opt_data, plog, inodir)
        output('Jacobian:')
        output(jac)
        output('Jacobian using finite differences:')
        output(jac_fd)
        output('Relative difference:')
        output((jac_fd - jac) / nm.where(jac_fd, jac_fd, 1.0))

        opt_data.jac = jac.copy()
        opt_data.jac_fd = jac_fd

    opt_data.n_jac += 1

    dt = opt_data.tjac.stop()
    output(f'jacobian evaluation done in {dt} s.')

    return jac

def parse_args(args=None):
    helps = {}

    sdefault_opt_conf = """ftol=1e-08, xtol=1e-4, gtol=None, method='trf',
                           f_scale=3e-8, loss='soft_l1', verbose=2
                        """
    opts = dict(
        par_names = ('density,young,poisson', 'parameters to be identified'),
        opt_conf = (sdefault_opt_conf, 'optimization solver options'),
        jac = (True, 'do not use the semi-analytical Jacobian calculation'),
        check_jac = (False,
                     'check the Jacobian using finite differences'),
        multi_rhs = (False, 'solve all rhs of sensitivity analysis in one call'),
        young = (200e9, """Young's modulus"""),
        poisson = (0.3, """Poisson's ratio"""),
        density = (7800.0, 'density'),
        alpha = (0.0, 'proportional damping coefficient (M)'),
        beta = (0.0, 'proportional damping coefficient (K)'),
        dims = ('1e-2,2.5e-3,2.5e-3',
                'physical dimensions of the block (L, d, x)'),
        shape = ('21,6,6',
                 'numbers of mesh vertices along each axis'),
        v0 = (1.0, 'initial impact velocity'),
        ct1 = (1.5, 'final time in L / "longitudinal wave speed" units'),
        dt = (None, 'time step (None means automatic)'),
        active_only = (False,
                       'vectors and matrices contain only active DOFs'),
        save_times = (20, 'number of time steps to save'),
        output_dir = ('output/edi', 'output directory'),
        plot_log = (False, 'show live log figures'),
        shell = (False, 'run ipython shell after all computations'),
        debug = (False,
                 'automatically start debugger when an exception is raised'),
    )

    parser = ArgumentParser(description=__doc__.rstrip(),
                            formatter_class=RawDescriptionHelpFormatter)

    dhelp = ' [default: %(default)s]'
    for key, (val, msg) in opts.items():
        helps[key] = msg
        action = 'store'
        vtype = type(val)
        choices = None
        option = key
        if val is True:
            action = 'store_false'
            option = 'no_' + key

        elif val is False:
            action = 'store_true'

        elif isinstance(val, tuple):
            choices = val
            vtype = type(val[0])
            val = val[0]

        elif isinstance(val, list):
            vtype = type(val[1])
            val = val[0]

        if action == 'store':
            helps[key] += dhelp
            parser.add_argument('--' + option.replace('_', '-'),
                                type=vtype,
                                action=action, dest=key, choices=choices,
                                default=val, help=helps[key])
        else:
            parser.add_argument('--' + option.replace('_', '-'),
                                action=action, dest=key,
                                default=val, help=helps[key])
    options = parser.parse_args(args=args)

    options.dims = [float(ii) for ii in options.dims.split(',')]
    options.shape = [int(ii) for ii in options.shape.split(',')]
    options.par_names = [ii.strip()  for ii in options.par_names.split(',')]
    aux = parse_as_dict(sdefault_opt_conf)
    aux.update(parse_as_dict(options.opt_conf))
    options.opt_conf = aux

    options.plane = 'strain'
    options.opt_bounds = {
        'density' : [0.5 * options.density, 1.5 * options.density],
        'young' : [0.5 * options.young, 1.5 * options.young],
        'poisson' : [0.9 * options.poisson, 1.1 * options.poisson],
        'alpha' : [0, 1e4],
        'beta' : [0, 1e-4],
    }

    return options, helps

def main():
    from scipy.optimize import least_squares

    from sfepy.base.base import output, Struct
    from sfepy.base.conf import ProblemConf
    from sfepy.discrete import Problem
    from sfepy.base.log import Log

    options, helps = parse_args()
    if options.debug:
        from sfepy.base.base import debug_on_error; debug_on_error()

    conf = ProblemConf.from_dict(define(**vars(options)), sys.modules[__name__])

    pb = Problem.from_conf(conf)

    out = []
    pb.solve(save_results=False, step_hook=partial(apply_sensor, out=out))
    usens = nm.concatenate(out)

    size = nm.abs(usens[:, -1]).max()
    true_data = usens[:, -1]
    data = true_data.copy()
    i0 = int(0.2 * options.shape[0])
    data[i0:] += nm.random.default_rng(seed=42).normal(
        0, 0.1 * size, len(data[i0:]),
    )
    # Add an outlier.
    data[max(len(data) // 2, i0)] += 0.5 * size

    par_info = {
        'density' : ('m', 'Omega', 'density', options.density),
        'young' : ('m', 'Omega', 'young', options.young * 0.8),
        'poisson' : ('m', 'Omega', 'poisson', options.poisson),
        'alpha' : ('term', 'eq', 1, 1000),
        'beta' : ('term', 'eq', 2, 1e-5),
    }
    par_names = options.par_names
    all_par_names = set(par_info.keys())
    if not set(par_names).issubset(all_par_names):
        unknown = set(par_names).difference(all_par_names)
        raise ValueError(f'parameters {unknown} not in {all_par_names}!')

    labels = [[name] for name in par_names] + [['|res|']]
    nlog = len(labels)
    formats = [['{:.16e}' for ii in range(len(labs))] for labs in labels]
    inodir = partial(op.join, options.output_dir)
    plog = Log(labels,
               formats=formats,
               yscales=['log'] * nlog,
               xlabels=['iter'] * nlog,
               ylabels=None,
               show_legends=True,
               is_plot=options.plot_log,
               log_filename=inodir('pars.txt'),
               aggregate=1000, sleep=0.1)

    pars0 = [par_info[name][-1] for name in par_names]
    pars0 = nm.array(pars0, dtype=nm.float64)
    x_scale = [par if abs(par) > 0 else 1.0 for par in pars0]
    bounds = [options.opt_bounds.get(name, (-nm.inf, nm.inf))
              for name in par_names]
    bounds = tuple(zip(*bounds))
    opt_data = Struct(name='opt_data', true_data=true_data, n_fun=0, n_jac=0,
                      tfun=Timer('fun'), tjac=Timer('jac'))

    timer = Timer(start=True)
    res = least_squares(
        eval_fun, pars0,
        jac=eval_jac if options.jac else '2-point',
        bounds=bounds,
        x_scale=x_scale,
        args=(data, pb, options, par_names, par_info, opt_data, plog, inodir),
        **options.opt_conf
    )
    elapsed = timer.stop()
    output(f'identification done in {elapsed} s.')
    output(res)
    od = opt_data
    tfi = od.tfun.total / od.n_fun
    tji = od.tjac.total / max(od.n_jac, 1)
    tfe = tfi * len(pars0)
    output(f'evaluations: function: {od.n_fun}, jacobian: {od.n_jac}')
    output(f'elapsed: function: {od.tfun.total}, jacobian {od.tjac.total}')
    output(f'elapsed/it: function: {tfi}, jacobian {tji}')
    output(f'expected finite difference jacobian time: {tfe}')

    plog(save_figure=inodir('pars.png'))

    if options.shell:
        from sfepy.base.base import shell; shell()

    plog(finished=True)

if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
Simple mesh generators and statistics.
"""
from argparse import ArgumentParser

from . import blockgen, cylindergen, gen_iga_patch, show_mesh_info

def main():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('--version', action='version', version='%(prog)s')
    subparsers = parser.add_subparsers(title='subcommands',
                                       description='valid subcommands',
                                       help='additional help',
                                       dest='mesh_kind',
                                       required=True)

    parser0 = subparsers.add_parser('block', help='generate a block mesh')
    parser0.set_defaults(fun=blockgen.gen_block)
    blockgen.add_args(parser0)

    parser1 = subparsers.add_parser('cylinder', help='generate a cylinder mesh')
    parser1.set_defaults(fun=cylindergen.gen_cylinder)
    cylindergen.add_args(parser1)

    parser2 = subparsers.add_parser('iga-patch', help='generate an IGA patch')
    parser2.set_defaults(fun=gen_iga_patch.gen_iga_patch)
    gen_iga_patch.add_args(parser2)

    parser3 = subparsers.add_parser('info', help='show various mesh statistics')
    parser3.set_defaults(fun=show_mesh_info.show_mesh_info)
    show_mesh_info.add_args(parser3)

    options = parser.parse_args()

    options.fun(options)

if __name__ == '__main__':
    main()



import os.path as op

import numpy as nm

from sfepy.base.conf import transform_variables
import sfepy.base.testing as tst

variables = {
    'u'       : ('unknown field', 'f', 0),
    'v'       : ('test field',    'f', 'u'),
}

def in_dir(adir):
    return lambda x: op.join(adir, x)

def gen_datas(meshes):
    datas = {}

    for key, mesh in meshes.items():
        bbox = mesh.get_bounding_box()
        nx = bbox[1,0] - bbox[0,0]
        centre = 0.5 * bbox.sum(axis=0)
        mesh.coors[:] -= centre

        data = nm.sin(4.0 * nm.pi * mesh.coors[:,0:1] / nx)
        datas['scalar_' + key] = data

        data = nm.zeros_like(mesh.coors)
        data[:,0] = 0.05 * nx * nm.sin(4.0 * nm.pi * mesh.coors[:,0] / nx)
        data[:,2] = 0.05 * nx * nm.cos(4.0 * nm.pi * mesh.coors[:,0] / nx)
        datas['vector_' + key] = data

    return datas

def do_interpolation(m2, m1, data, field_name, force=False):
    """Interpolate data from m1 to m2. """
    from sfepy.discrete import Variables
    from sfepy.discrete.fem import FEDomain, Field

    fields = {
        'scalar_si' : ((1,1), 'Omega', 2),
        'vector_si' : ((3,1), 'Omega', 2),
        'scalar_tp' : ((1,1), 'Omega', 1),
        'vector_tp' : ((3,1), 'Omega', 1),
    }

    d1 = FEDomain('d1', m1)

    d1.create_region('Omega', 'all')

    f = fields[field_name]

    field1 = Field.from_args('f', nm.float64, f[0], d1.regions[f[1]],
                             approx_order=f[2])
    ff = {field1.name : field1}

    vv = Variables.from_conf(transform_variables(variables), ff)
    u1 = vv['u']
    u1.set_from_mesh_vertices(data)

    d2 = FEDomain('d2', m2)
    d2.create_region('Omega', 'all')

    field2 = Field.from_args('f', nm.float64, f[0], d2.regions[f[1]],
                             approx_order=f[2])
    ff2 = {field2.name : field2}

    vv2 = Variables.from_conf(transform_variables(variables), ff2)
    u2 = vv2['u']

    if not force:
        # Performs interpolation, if other field differs from self.field
        # or, in particular, is defined on a different mesh.
        u2.set_from_other(u1, strategy='interpolation', close_limit=0.5)

    else:
        coors = u2.field.get_coor()
        vals = u1.evaluate_at(coors, close_limit=0.5)
        u2.set_data(vals)

    return u1, u2

def prepare_variable(filename, n_components):
    from sfepy.discrete import FieldVariable
    from sfepy.discrete.fem import Mesh, FEDomain, Field

    mesh = Mesh.from_file(filename)

    bbox = mesh.get_bounding_box()
    dd = bbox[1,:] - bbox[0,:]
    data = (nm.sin(4.0 * nm.pi * mesh.coors[:,0:1] / dd[0])
            * nm.cos(4.0 * nm.pi * mesh.coors[:,1:2] / dd[1]))

    domain = FEDomain('domain', mesh)
    omega = domain.create_region('Omega', 'all')
    field = Field.from_args('field', nm.float64, n_components, omega,
                            approx_order=2)

    u = FieldVariable('u', 'parameter', field,
                      primary_var_name='(set-to-None)')
    u.set_from_mesh_vertices(data * nm.arange(1, n_components + 1)[None, :])

    return u

def test_interpolation(output_dir):
    from sfepy import data_dir
    from sfepy.discrete.fem import Mesh
    from sfepy.linalg import make_axis_rotation_matrix

    fname = in_dir(output_dir)

    meshes = {
        'tp' : Mesh.from_file(data_dir + '/meshes/3d/block.mesh'),
        'si' : Mesh.from_file(data_dir + '/meshes/3d/cylinder.mesh'),
    }

    datas = gen_datas(meshes)

    for field_name in ['scalar_si', 'vector_si', 'scalar_tp', 'vector_tp']:
        m1 = meshes[field_name[-2:]]
        for ia, angle in enumerate(nm.linspace(0.0, nm.pi, 11)):
            tst.report('%s: %d. angle: %f' % (field_name, ia, angle))
            mtx = make_axis_rotation_matrix([0, 1, 0], angle)

            m2 = m1.copy('rotated mesh')
            m2.transform_coors(mtx)

            data = datas[field_name]
            u1, u2 = do_interpolation(m2, m1, data, field_name)

            if ia == 0:
                u1.save_as_mesh(fname('test_mesh_interp_%s_u1.vtk'
                                      % field_name))

            u2.save_as_mesh(fname('test_mesh_interp_%s_u2.%03d.vtk'
                                  % (field_name, ia)))

def test_interpolation_two_meshes(output_dir):
    from sfepy import data_dir
    from sfepy.discrete import Variables
    from sfepy.discrete.fem import Mesh, FEDomain, Field

    m1 = Mesh.from_file(data_dir + '/meshes/3d/block.mesh')

    m2 = Mesh.from_file(data_dir + '/meshes/3d/cube_medium_tetra.mesh')
    m2.coors[:] *= 2.0

    bbox = m1.get_bounding_box()
    dd = bbox[1,:] - bbox[0,:]
    data = nm.sin(4.0 * nm.pi * m1.coors[:,0:1] / dd[0]) \
           * nm.cos(4.0 * nm.pi * m1.coors[:,1:2] / dd[1])

    variables1 = {
        'u'       : ('unknown field', 'scalar_tp', 0),
        'v'       : ('test field',    'scalar_tp', 'u'),
    }

    variables2 = {
        'u'       : ('unknown field', 'scalar_si', 0),
        'v'       : ('test field',    'scalar_si', 'u'),
    }

    d1 = FEDomain('d1', m1)
    omega1 = d1.create_region('Omega', 'all')
    field1 = Field.from_args('scalar_tp', nm.float64, (1,1), omega1,
                             approx_order=1)
    ff1 = {field1.name : field1}

    d2 = FEDomain('d2', m2)
    omega2 = d2.create_region('Omega', 'all')
    field2 = Field.from_args('scalar_si', nm.float64, (1,1), omega2,
                             approx_order=0)
    ff2 = {field2.name : field2}

    vv1 = Variables.from_conf(transform_variables(variables1), ff1)
    u1 = vv1['u']
    u1.set_from_mesh_vertices(data)

    vv2 = Variables.from_conf(transform_variables(variables2), ff2)
    u2 = vv2['u']

    # Performs interpolation, if other field differs from self.field
    # or, in particular, is defined on a different mesh.
    u2.set_from_other(u1, strategy='interpolation', close_limit=0.1)

    fname = in_dir(output_dir)
    u1.save_as_mesh(fname('test_mesh_interp_block_scalar.vtk'))
    u2.save_as_mesh(fname('test_mesh_interp_cube_scalar.vtk'))

def test_invariance():
    from sfepy import data_dir
    from sfepy.discrete.fem import Mesh

    meshes = {
        'tp' : Mesh.from_file(data_dir + '/meshes/3d/block.mesh'),
        'si' : Mesh.from_file(data_dir + '/meshes/3d/cylinder.mesh'),
    }
    datas = gen_datas(meshes)

    ok = True
    for field_name in ['scalar_si', 'vector_si', 'scalar_tp', 'vector_tp']:
        m1 = meshes[field_name[-2:]]

        data = datas[field_name]
        u1, u2 = do_interpolation(m1, m1, data, field_name, force=True)

        tst.report('max. difference:', nm.abs(u1() - u2()).max())
        _ok = nm.allclose(u1(), u2(), rtol=0.0, atol=1e-12)
        tst.report('invariance for %s field: %s' % (field_name, _ok))

        ok = ok and _ok

    assert ok

def test_invariance_qp():
    from sfepy import data_dir
    from sfepy.discrete import Integral
    from sfepy.terms import Term
    from sfepy.discrete.common.mappings import get_physical_qps

    ok = True
    for name in ['meshes/3d/block.mesh', 'meshes/3d/cylinder.mesh',
                 'meshes/2d/square_quad.mesh',
                 'meshes/2d/square_unit_tri.mesh']:
        tst.report(name)

        u = prepare_variable(op.join(data_dir, name), n_components=3)
        omega = u.field.region

        integral = Integral('i', order=3)
        qps = get_physical_qps(omega, integral)
        coors = qps.values

        term = Term.new('ev_integrate(u)', integral, omega, u=u)
        term.setup()
        val1 = term.evaluate(mode='qp')
        val1 = val1.ravel()

        val2 = u.evaluate_at(coors).ravel()

        tst.report('value: max. difference:', nm.abs(val1 - val2).max())
        ok1 = nm.allclose(val1, val2, rtol=0.0, atol=1e-12)
        tst.report('->', ok1)

        term = Term.new('ev_grad(u)', integral, omega, u=u)
        term.setup()
        val1 = term.evaluate(mode='qp')
        val1 = val1.transpose((0, 1, 3, 2)).ravel()

        val2 = u.evaluate_at(coors, mode='grad').ravel()

        tst.report('gradient: max. difference:', nm.abs(val1 - val2).max())
        ok2 = nm.allclose(val1, val2, rtol=0.0, atol=1e-10)
        tst.report('->', ok2)

        ok = ok and ok1 and ok2

    assert ok

def test_field_gradient():
    from sfepy import data_dir

    ok = True
    for name in ['meshes/3d/block.mesh', 'meshes/3d/cylinder.mesh',
                 'meshes/2d/square_quad.mesh',
                 'meshes/2d/square_unit_tri.mesh']:
        tst.report(name)

        u = prepare_variable(op.join(data_dir, name), n_components=5)

        bbox = u.field.domain.get_mesh_bounding_box()
        coors = nm.c_[tuple([nm.linspace(ii[0] + 1e-3, ii[1] - 1e-3, 100)
                             for ii in bbox.T])]

        grad, cells, status = u.evaluate_at(coors, mode='grad',
                                            close_limit=0.0,
                                            ret_status=True)
        agrad = nm.dot(grad[:, :, :], nm.ones((grad.shape[2], 1)))[..., 0]

        eps = 1e-5
        val0 = u.evaluate_at(coors - eps, close_limit=0.0)[..., 0]
        val1 = u.evaluate_at(coors + eps, close_limit=0.0)[..., 0]

        ngrad = 0.5 * (val1 - val0) / eps

        ii = nm.where(status == 0)

        tst.report('max. difference:', nm.abs(agrad[ii] - ngrad[ii]).max())

        _ok = nm.allclose(agrad[ii], ngrad[ii], rtol=0.0, atol=10 * eps)
        tst.report('->', _ok)

        ok = ok and _ok

        for ic in range(1, u.n_components):
            _ok = nm.allclose((ic + 1) * agrad[ii, 0], agrad[ii, ic],
                              rtol=0.0, atol=1e-12)
            tst.report('component %d / component 0: mean: %.2f'
                        % (ic, (agrad[ii, ic] / agrad[ii, 0]).mean()))
            tst.report('->', _ok)

            ok = ok and _ok

    assert ok

def test_evaluate_at():
    from sfepy import data_dir
    from sfepy.discrete.fem import Mesh
    from sfepy.discrete import Variables
    from sfepy.discrete.fem import FEDomain, Field

    meshes = {
        'tp' : Mesh.from_file(data_dir + '/meshes/3d/block.mesh'),
    }
    datas = gen_datas(meshes)

    fields = {
        'scalar_tp' : ((1,1), 'Omega', 1),
        'vector_tp' : ((3,1), 'Omega', 1),
    }

    ok = True
    for field_name in ['scalar_tp', 'vector_tp']:
        d = FEDomain('d', meshes['tp'])
        d.create_region('Omega', 'all')

        f = fields[field_name]
        field = Field.from_args('f', nm.complex128, f[0],
                                d.regions[f[1]],
                                approx_order=f[2])
        ff = {field.name : field}

        vv = Variables.from_conf(transform_variables(variables), ff)
        u = vv['u']

        bbox = d.get_mesh_bounding_box()
        t = nm.expand_dims(nm.linspace(0, 1, 100), 1)
        coors = nm.expand_dims(bbox[1] - bbox[0], 0) * t + bbox[0]

        data_r = datas[field_name]
        data_i = 2. / (1 + datas[field_name])

        u.set_from_mesh_vertices(data_r)
        vals_r = u.evaluate_at(coors)
        u.set_from_mesh_vertices(data_i)
        vals_i = u.evaluate_at(coors)
        u.set_from_mesh_vertices(data_r + data_i * 1j)
        vals = u.evaluate_at(coors)

        _ok = nm.allclose(vals_r + vals_i * 1j, vals, rtol=0.0, atol=1e-12)
        _ok = _ok and nm.abs(vals).sum() > 1
        tst.report('evaluating complex field %s: %s' % (field_name, _ok))

        ok = ok and _ok

    assert ok
